from typing import Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from aeromaps.models.base import AeromapsModel


class PassengerAircraftDocNonEnergyComplex(AeromapsModel):
    def __init__(
        self, name="passenger_aircraft_doc_non_energy_complex", fleet_model=None, *args, **kwargs
    ):
        super().__init__(name=name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        dummy_fleet_model_output: np.ndarray,
        ask_long_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range: pd.Series = pd.Series(dtype="float64"),
        ask_short_range: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:

        doc_non_energy_per_ask_short_range_dropin_fuel = self.fleet_model.df[
            "Short Range:doc_non_energy:dropin_fuel"
        ]
        doc_non_energy_per_ask_medium_range_dropin_fuel = self.fleet_model.df[
            "Medium Range:doc_non_energy:dropin_fuel"
        ]
        doc_non_energy_per_ask_long_range_dropin_fuel = self.fleet_model.df[
            "Long Range:doc_non_energy:dropin_fuel"
        ]
        doc_non_energy_per_ask_short_range_hydrogen = self.fleet_model.df[
            "Short Range:doc_non_energy:hydrogen"
        ]
        doc_non_energy_per_ask_medium_range_hydrogen = self.fleet_model.df[
            "Medium Range:doc_non_energy:hydrogen"
        ]
        doc_non_energy_per_ask_long_range_hydrogen = self.fleet_model.df[
            "Long Range:doc_non_energy:hydrogen"
        ]

        # Drop-in - Projections

        self.df.loc[
            :, "doc_non_energy_per_ask_short_range_dropin_fuel"
        ] = doc_non_energy_per_ask_short_range_dropin_fuel
        self.df.loc[
            :, "doc_non_energy_per_ask_medium_range_dropin_fuel"
        ] = doc_non_energy_per_ask_medium_range_dropin_fuel
        self.df.loc[
            :, "doc_non_energy_per_ask_long_range_dropin_fuel"
        ] = doc_non_energy_per_ask_long_range_dropin_fuel

        # Hydrogen
        self.df.loc[
            :, "doc_non_energy_per_ask_short_range_hydrogen"
        ] = doc_non_energy_per_ask_short_range_hydrogen
        self.df.loc[
            :, "doc_non_energy_per_ask_medium_range_hydrogen"
        ] = doc_non_energy_per_ask_medium_range_hydrogen
        self.df.loc[
            :, "doc_non_energy_per_ask_long_range_hydrogen"
        ] = doc_non_energy_per_ask_long_range_hydrogen

        doc_non_energy_per_ask_long_range_mean = (
            doc_non_energy_per_ask_long_range_hydrogen * ask_long_range_hydrogen_share / 100
            + doc_non_energy_per_ask_long_range_dropin_fuel * ask_long_range_dropin_fuel_share / 100
        )

        doc_non_energy_per_ask_medium_range_mean = (
            doc_non_energy_per_ask_medium_range_hydrogen * ask_medium_range_hydrogen_share / 100
            + doc_non_energy_per_ask_medium_range_dropin_fuel
            * ask_medium_range_dropin_fuel_share
            / 100
        )

        doc_non_energy_per_ask_short_range_mean = (
            doc_non_energy_per_ask_short_range_hydrogen * ask_short_range_hydrogen_share / 100
            + doc_non_energy_per_ask_short_range_dropin_fuel
            * ask_short_range_dropin_fuel_share
            / 100
        )

        doc_non_energy_per_ask_mean = (
            doc_non_energy_per_ask_long_range_mean * ask_long_range
            + doc_non_energy_per_ask_medium_range_mean * ask_medium_range
            + doc_non_energy_per_ask_short_range_mean * ask_short_range
        ) / (ask_long_range + ask_medium_range + ask_short_range)

        self.df.loc[
            :, "doc_non_energy_per_ask_long_range_mean"
        ] = doc_non_energy_per_ask_long_range_mean

        self.df.loc[
            :, "doc_non_energy_per_ask_medium_range_mean"
        ] = doc_non_energy_per_ask_medium_range_mean

        self.df.loc[
            :, "doc_non_energy_per_ask_short_range_mean"
        ] = doc_non_energy_per_ask_short_range_mean

        self.df.loc[:, "doc_non_energy_per_ask_mean"] = doc_non_energy_per_ask_mean

        return (
            doc_non_energy_per_ask_short_range_dropin_fuel,
            doc_non_energy_per_ask_medium_range_dropin_fuel,
            doc_non_energy_per_ask_long_range_dropin_fuel,
            doc_non_energy_per_ask_short_range_hydrogen,
            doc_non_energy_per_ask_medium_range_hydrogen,
            doc_non_energy_per_ask_long_range_hydrogen,
            doc_non_energy_per_ask_short_range_mean,
            doc_non_energy_per_ask_medium_range_mean,
            doc_non_energy_per_ask_long_range_mean,
            doc_non_energy_per_ask_mean,
        )


class PassengerAircraftDocNonEnergySimple(AeromapsModel):
    def __init__(self, name="passenger_aircraft_doc_non_energy_simple", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        doc_non_energy_per_ask_short_range_dropin_fuel_init: float = 0.0,
        doc_non_energy_per_ask_medium_range_dropin_fuel_init: float = 0.0,
        doc_non_energy_per_ask_long_range_dropin_fuel_init: float = 0.0,
        doc_non_energy_per_ask_short_range_dropin_fuel_gain: float = 0.0,
        doc_non_energy_per_ask_medium_range_dropin_fuel_gain: float = 0.0,
        doc_non_energy_per_ask_long_range_dropin_fuel_gain: float = 0.0,
        relative_doc_non_energy_per_ask_hydrogen_wrt_dropin_short_range: float = 0.0,
        relative_doc_non_energy_per_ask_hydrogen_wrt_dropin_medium_range: float = 0.0,
        relative_doc_non_energy_per_ask_hydrogen_wrt_dropin_long_range: float = 0.0,
        ask_long_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range: pd.Series = pd.Series(dtype="float64"),
        ask_short_range: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """DOC (without energy DOC) per ASK calculation using simple models."""

        # Initialization based on 2019 values

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[
                k, "doc_non_energy_per_ask_short_range_dropin_fuel"
            ] = doc_non_energy_per_ask_short_range_dropin_fuel_init
            self.df.loc[
                k, "doc_non_energy_per_ask_medium_range_dropin_fuel"
            ] = doc_non_energy_per_ask_medium_range_dropin_fuel_init
            self.df.loc[
                k, "doc_non_energy_per_ask_long_range_dropin_fuel"
            ] = doc_non_energy_per_ask_long_range_dropin_fuel_init

        # Projections

        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "doc_non_energy_per_ask_short_range_dropin_fuel"] = self.df.loc[
                k - 1, "doc_non_energy_per_ask_short_range_dropin_fuel"
            ] * (1 - doc_non_energy_per_ask_short_range_dropin_fuel_gain / 100)
            self.df.loc[k, "doc_non_energy_per_ask_medium_range_dropin_fuel"] = self.df.loc[
                k - 1, "doc_non_energy_per_ask_medium_range_dropin_fuel"
            ] * (1 - doc_non_energy_per_ask_medium_range_dropin_fuel_gain / 100)
            self.df.loc[k, "doc_non_energy_per_ask_long_range_dropin_fuel"] = self.df.loc[
                k - 1, "doc_non_energy_per_ask_long_range_dropin_fuel"
            ] * (1 - doc_non_energy_per_ask_long_range_dropin_fuel_gain / 100)

        doc_non_energy_per_ask_short_range_dropin_fuel = self.df[
            "doc_non_energy_per_ask_short_range_dropin_fuel"
        ]
        doc_non_energy_per_ask_medium_range_dropin_fuel = self.df[
            "doc_non_energy_per_ask_medium_range_dropin_fuel"
        ]
        doc_non_energy_per_ask_long_range_dropin_fuel = self.df[
            "doc_non_energy_per_ask_long_range_dropin_fuel"
        ]

        doc_non_energy_per_ask_short_range_hydrogen = (
            doc_non_energy_per_ask_short_range_dropin_fuel
            * relative_doc_non_energy_per_ask_hydrogen_wrt_dropin_short_range
        )
        doc_non_energy_per_ask_medium_range_hydrogen = (
            doc_non_energy_per_ask_medium_range_dropin_fuel
            * relative_doc_non_energy_per_ask_hydrogen_wrt_dropin_medium_range
        )
        doc_non_energy_per_ask_long_range_hydrogen = (
            doc_non_energy_per_ask_long_range_dropin_fuel
            * relative_doc_non_energy_per_ask_hydrogen_wrt_dropin_long_range
        )

        self.df.loc[
            :, "doc_non_energy_per_ask_short_range_hydrogen"
        ] = doc_non_energy_per_ask_short_range_hydrogen
        self.df.loc[
            :, "doc_non_energy_per_ask_medium_range_hydrogen"
        ] = doc_non_energy_per_ask_medium_range_hydrogen
        self.df.loc[
            :, "doc_non_energy_per_ask_long_range_hydrogen"
        ] = doc_non_energy_per_ask_long_range_hydrogen

        doc_non_energy_per_ask_long_range_mean = (
            doc_non_energy_per_ask_long_range_hydrogen * ask_long_range_hydrogen_share / 100
            + doc_non_energy_per_ask_long_range_dropin_fuel * ask_long_range_dropin_fuel_share / 100
        )

        doc_non_energy_per_ask_medium_range_mean = (
            doc_non_energy_per_ask_medium_range_hydrogen * ask_medium_range_hydrogen_share / 100
            + doc_non_energy_per_ask_medium_range_dropin_fuel
            * ask_medium_range_dropin_fuel_share
            / 100
        )

        doc_non_energy_per_ask_short_range_mean = (
            doc_non_energy_per_ask_short_range_hydrogen * ask_short_range_hydrogen_share / 100
            + doc_non_energy_per_ask_short_range_dropin_fuel
            * ask_short_range_dropin_fuel_share
            / 100
        )

        doc_non_energy_per_ask_mean = (
            doc_non_energy_per_ask_long_range_mean * ask_long_range
            + doc_non_energy_per_ask_medium_range_mean * ask_medium_range
            + doc_non_energy_per_ask_short_range_mean * ask_short_range
        ) / (ask_long_range + ask_medium_range + ask_short_range)

        self.df.loc[
            :, "doc_non_energy_per_ask_long_range_mean"
        ] = doc_non_energy_per_ask_long_range_mean

        self.df.loc[
            :, "doc_non_energy_per_ask_medium_range_mean"
        ] = doc_non_energy_per_ask_medium_range_mean

        self.df.loc[
            :, "doc_non_energy_per_ask_short_range_mean"
        ] = doc_non_energy_per_ask_short_range_mean

        self.df.loc[:, "doc_non_energy_per_ask_mean"] = doc_non_energy_per_ask_mean

        return (
            doc_non_energy_per_ask_short_range_dropin_fuel,
            doc_non_energy_per_ask_medium_range_dropin_fuel,
            doc_non_energy_per_ask_long_range_dropin_fuel,
            doc_non_energy_per_ask_short_range_hydrogen,
            doc_non_energy_per_ask_medium_range_hydrogen,
            doc_non_energy_per_ask_long_range_hydrogen,
            doc_non_energy_per_ask_short_range_mean,
            doc_non_energy_per_ask_medium_range_mean,
            doc_non_energy_per_ask_long_range_mean,
            doc_non_energy_per_ask_mean,
        )


class PassengerAircraftDocEnergy(AeromapsModel):
    def __init__(self, name="passenger_aircraft_doc_energy", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        energy_per_ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_long_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_medium_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_short_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        dropin_mean_mfsp: pd.Series = pd.Series(dtype="float64"),
        h2_avg_cost_per_kg: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range: pd.Series = pd.Series(dtype="float64"),
        ask_short_range: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        # Drop-in fuels lower heating value (MJ/L)
        fuel_lhv = 35.3
        # LH2 specific energy (MJ/kg)
        hydrogen_specific_energy = 119.93

        doc_energy_per_ask_long_range_dropin_fuel = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_energy_per_ask_long_range_hydrogen = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_energy_per_ask_medium_range_dropin_fuel = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_energy_per_ask_medium_range_hydrogen = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_energy_per_ask_short_range_dropin_fuel = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_energy_per_ask_short_range_hydrogen = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        for k in range(self.historic_start_year, self.end_year + 1):
            if ask_long_range_dropin_fuel_share[k] > 0:
                doc_energy_per_ask_long_range_dropin_fuel[k] = (
                    energy_per_ask_long_range_dropin_fuel[k] * dropin_mean_mfsp[k] / fuel_lhv
                )
            if ask_long_range_hydrogen_share[k] > 0:
                doc_energy_per_ask_long_range_hydrogen[k] = (
                    energy_per_ask_long_range_hydrogen[k]
                    * h2_avg_cost_per_kg[k]
                    / hydrogen_specific_energy
                )
            if ask_medium_range_dropin_fuel_share[k] > 0:
                doc_energy_per_ask_medium_range_dropin_fuel[k] = (
                    energy_per_ask_medium_range_dropin_fuel[k] * dropin_mean_mfsp[k] / fuel_lhv
                )
            if ask_medium_range_hydrogen_share[k] > 0:
                doc_energy_per_ask_medium_range_hydrogen[k] = (
                    energy_per_ask_medium_range_hydrogen[k]
                    * h2_avg_cost_per_kg[k]
                    / hydrogen_specific_energy
                )
            if ask_short_range_dropin_fuel_share[k] > 0:
                doc_energy_per_ask_short_range_dropin_fuel[k] = (
                    energy_per_ask_short_range_dropin_fuel[k] * dropin_mean_mfsp[k] / fuel_lhv
                )
            if ask_short_range_hydrogen_share[k] > 0:
                doc_energy_per_ask_short_range_hydrogen[k] = (
                    energy_per_ask_short_range_hydrogen[k]
                    * h2_avg_cost_per_kg[k]
                    / hydrogen_specific_energy
                )

        doc_energy_per_ask_long_range_mean = (
            doc_energy_per_ask_long_range_hydrogen.fillna(0) * ask_long_range_hydrogen_share / 100
            + doc_energy_per_ask_long_range_dropin_fuel.fillna(0)
            * ask_long_range_dropin_fuel_share
            / 100
        )

        doc_energy_per_ask_medium_range_mean = (
            doc_energy_per_ask_medium_range_hydrogen.fillna(0)
            * ask_medium_range_hydrogen_share
            / 100
            + doc_energy_per_ask_medium_range_dropin_fuel.fillna(0)
            * ask_medium_range_dropin_fuel_share
            / 100
        )

        doc_energy_per_ask_short_range_mean = (
            doc_energy_per_ask_short_range_hydrogen.fillna(0) * ask_short_range_hydrogen_share / 100
            + doc_energy_per_ask_short_range_dropin_fuel.fillna(0)
            * ask_short_range_dropin_fuel_share
            / 100
        )

        doc_energy_per_ask_mean = (
            doc_energy_per_ask_long_range_mean * ask_long_range
            + doc_energy_per_ask_medium_range_mean * ask_medium_range
            + doc_energy_per_ask_short_range_mean * ask_short_range
        ) / (ask_long_range + ask_medium_range + ask_short_range)

        self.df.loc[
            :, "doc_energy_per_ask_long_range_dropin_fuel"
        ] = doc_energy_per_ask_long_range_dropin_fuel
        self.df.loc[
            :, "doc_energy_per_ask_long_range_hydrogen"
        ] = doc_energy_per_ask_long_range_hydrogen
        self.df.loc[:, "doc_energy_per_ask_long_range_mean"] = doc_energy_per_ask_long_range_mean
        self.df.loc[
            :, "doc_energy_per_ask_medium_range_dropin_fuel"
        ] = doc_energy_per_ask_medium_range_dropin_fuel
        self.df.loc[
            :, "doc_energy_per_ask_medium_range_hydrogen"
        ] = doc_energy_per_ask_medium_range_hydrogen
        self.df.loc[
            :, "doc_energy_per_ask_medium_range_mean"
        ] = doc_energy_per_ask_medium_range_mean
        self.df.loc[
            :, "doc_energy_per_ask_short_range_dropin_fuel"
        ] = doc_energy_per_ask_short_range_dropin_fuel
        self.df.loc[
            :, "doc_energy_per_ask_short_range_hydrogen"
        ] = doc_energy_per_ask_short_range_hydrogen
        self.df.loc[:, "doc_energy_per_ask_short_range_mean"] = doc_energy_per_ask_short_range_mean
        self.df.loc[:, "doc_energy_per_ask_mean"] = doc_energy_per_ask_mean
        return (
            doc_energy_per_ask_long_range_dropin_fuel,
            doc_energy_per_ask_long_range_hydrogen,
            doc_energy_per_ask_long_range_mean,
            doc_energy_per_ask_medium_range_dropin_fuel,
            doc_energy_per_ask_medium_range_hydrogen,
            doc_energy_per_ask_medium_range_mean,
            doc_energy_per_ask_short_range_dropin_fuel,
            doc_energy_per_ask_short_range_hydrogen,
            doc_energy_per_ask_short_range_mean,
            doc_energy_per_ask_mean,
        )


class PassengerAircraftDocCarbonTax(AeromapsModel):
    def __init__(self, name="passenger_aircraft_doc_carbon_tax", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        energy_per_ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_long_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_medium_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        energy_per_ask_short_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        dropin_mfsp_carbon_tax_supplement: pd.Series = pd.Series(dtype="float64"),
        h2_avg_carbon_tax_per_kg: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_hydrogen_share: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_dropin_fuel_share: pd.Series = pd.Series(dtype="float64"),
        ask_long_range: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range: pd.Series = pd.Series(dtype="float64"),
        ask_short_range: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        # Drop-in fuels lower heating value (MJ/L)
        fuel_lhv = 35.3
        # LH2 specific energy (MJ/kg)
        hydrogen_specific_energy = 119.93

        doc_carbon_tax_per_ask_long_range_dropin_fuel = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_carbon_tax_per_ask_long_range_hydrogen = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_carbon_tax_per_ask_medium_range_dropin_fuel = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_carbon_tax_per_ask_medium_range_hydrogen = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_carbon_tax_per_ask_short_range_dropin_fuel = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        doc_carbon_tax_per_ask_short_range_hydrogen = pd.Series(
            np.nan, range(self.historic_start_year, self.end_year + 1)
        )
        for k in range(self.historic_start_year, self.end_year + 1):
            if ask_long_range_dropin_fuel_share[k] > 0:
                doc_carbon_tax_per_ask_long_range_dropin_fuel[k] = (
                    energy_per_ask_long_range_dropin_fuel[k]
                    * dropin_mfsp_carbon_tax_supplement[k]
                    / fuel_lhv
                )
            if ask_long_range_hydrogen_share[k] > 0:
                doc_carbon_tax_per_ask_long_range_hydrogen[k] = (
                    energy_per_ask_long_range_hydrogen[k]
                    * h2_avg_carbon_tax_per_kg[k]
                    / hydrogen_specific_energy
                )
            if ask_medium_range_dropin_fuel_share[k] > 0:
                doc_carbon_tax_per_ask_medium_range_dropin_fuel[k] = (
                    energy_per_ask_medium_range_dropin_fuel[k]
                    * dropin_mfsp_carbon_tax_supplement[k]
                    / fuel_lhv
                )
            if ask_medium_range_hydrogen_share[k] > 0:
                doc_carbon_tax_per_ask_medium_range_hydrogen[k] = (
                    energy_per_ask_medium_range_hydrogen[k]
                    * h2_avg_carbon_tax_per_kg[k]
                    / hydrogen_specific_energy
                )
            if ask_short_range_dropin_fuel_share[k] > 0:
                doc_carbon_tax_per_ask_short_range_dropin_fuel[k] = (
                    energy_per_ask_short_range_dropin_fuel[k]
                    * dropin_mfsp_carbon_tax_supplement[k]
                    / fuel_lhv
                )
            if ask_short_range_hydrogen_share[k] > 0:
                doc_carbon_tax_per_ask_short_range_hydrogen[k] = (
                    energy_per_ask_short_range_hydrogen[k]
                    * h2_avg_carbon_tax_per_kg[k]
                    / hydrogen_specific_energy
                )

        doc_carbon_tax_per_ask_long_range_mean = (
            doc_carbon_tax_per_ask_long_range_hydrogen.fillna(0)
            * ask_long_range_hydrogen_share
            / 100
            + doc_carbon_tax_per_ask_long_range_dropin_fuel.fillna(0)
            * ask_long_range_dropin_fuel_share
            / 100
        )

        doc_carbon_tax_per_ask_medium_range_mean = (
            doc_carbon_tax_per_ask_medium_range_hydrogen.fillna(0)
            * ask_medium_range_hydrogen_share
            / 100
            + doc_carbon_tax_per_ask_medium_range_dropin_fuel.fillna(0)
            * ask_medium_range_dropin_fuel_share
            / 100
        )

        doc_carbon_tax_per_ask_short_range_mean = (
            doc_carbon_tax_per_ask_short_range_hydrogen.fillna(0)
            * ask_short_range_hydrogen_share
            / 100
            + doc_carbon_tax_per_ask_short_range_dropin_fuel.fillna(0)
            * ask_short_range_dropin_fuel_share
            / 100
        )

        doc_carbon_tax_per_ask_mean = (
            doc_carbon_tax_per_ask_long_range_mean * ask_long_range
            + doc_carbon_tax_per_ask_medium_range_mean * ask_medium_range
            + doc_carbon_tax_per_ask_short_range_mean * ask_short_range
        ) / (ask_long_range + ask_medium_range + ask_short_range)

        self.df.loc[
            :, "doc_carbon_tax_per_ask_long_range_dropin_fuel"
        ] = doc_carbon_tax_per_ask_long_range_dropin_fuel
        self.df.loc[
            :, "doc_carbon_tax_per_ask_long_range_hydrogen"
        ] = doc_carbon_tax_per_ask_long_range_hydrogen
        self.df.loc[
            :, "doc_carbon_tax_per_ask_long_range_mean"
        ] = doc_carbon_tax_per_ask_long_range_mean
        self.df.loc[
            :, "doc_carbon_tax_per_ask_medium_range_dropin_fuel"
        ] = doc_carbon_tax_per_ask_medium_range_dropin_fuel
        self.df.loc[
            :, "doc_carbon_tax_per_ask_medium_range_hydrogen"
        ] = doc_carbon_tax_per_ask_medium_range_hydrogen
        self.df.loc[
            :, "doc_carbon_tax_per_ask_medium_range_mean"
        ] = doc_carbon_tax_per_ask_medium_range_mean
        self.df.loc[
            :, "doc_carbon_tax_per_ask_short_range_dropin_fuel"
        ] = doc_carbon_tax_per_ask_short_range_dropin_fuel
        self.df.loc[
            :, "doc_carbon_tax_per_ask_short_range_hydrogen"
        ] = doc_carbon_tax_per_ask_short_range_hydrogen
        self.df.loc[
            :, "doc_carbon_tax_per_ask_short_range_mean"
        ] = doc_carbon_tax_per_ask_short_range_mean
        self.df.loc[:, "doc_carbon_tax_per_ask_mean"] = doc_carbon_tax_per_ask_mean

        return (
            doc_carbon_tax_per_ask_long_range_dropin_fuel,
            doc_carbon_tax_per_ask_long_range_hydrogen,
            doc_carbon_tax_per_ask_long_range_mean,
            doc_carbon_tax_per_ask_medium_range_dropin_fuel,
            doc_carbon_tax_per_ask_medium_range_hydrogen,
            doc_carbon_tax_per_ask_medium_range_mean,
            doc_carbon_tax_per_ask_short_range_dropin_fuel,
            doc_carbon_tax_per_ask_short_range_hydrogen,
            doc_carbon_tax_per_ask_short_range_mean,
            doc_carbon_tax_per_ask_mean,
        )


class PassengerAircraftTotalDoc(AeromapsModel):
    def __init__(self, name="passenger_aircraft_total_doc", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        doc_non_energy_per_ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_short_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_medium_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_long_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_short_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_medium_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_long_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_non_energy_per_ask_mean: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_long_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_long_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_medium_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_medium_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_short_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_short_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_energy_per_ask_mean: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_long_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_long_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_medium_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_medium_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_short_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_short_range_mean: pd.Series = pd.Series(dtype="float64"),
        doc_carbon_tax_per_ask_mean: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        # dropin
        doc_total_per_ask_short_range_dropin_fuel = (
            doc_non_energy_per_ask_short_range_dropin_fuel
            + doc_energy_per_ask_short_range_dropin_fuel
            + doc_carbon_tax_per_ask_short_range_dropin_fuel
        )

        doc_total_per_ask_medium_range_dropin_fuel = (
            doc_non_energy_per_ask_medium_range_dropin_fuel
            + doc_energy_per_ask_medium_range_dropin_fuel
            + doc_carbon_tax_per_ask_medium_range_dropin_fuel
        )

        doc_total_per_ask_long_range_dropin_fuel = (
            doc_non_energy_per_ask_long_range_dropin_fuel
            + doc_energy_per_ask_long_range_dropin_fuel
            + doc_carbon_tax_per_ask_long_range_dropin_fuel
        )

        # Hydrogen
        doc_total_per_ask_short_range_hydrogen = (
            doc_non_energy_per_ask_short_range_hydrogen
            + doc_energy_per_ask_short_range_hydrogen
            + doc_carbon_tax_per_ask_short_range_hydrogen
        )

        doc_total_per_ask_medium_range_hydrogen = (
            doc_non_energy_per_ask_medium_range_hydrogen
            + doc_energy_per_ask_medium_range_hydrogen
            + doc_carbon_tax_per_ask_medium_range_hydrogen
        )

        doc_total_per_ask_long_range_hydrogen = (
            doc_non_energy_per_ask_long_range_hydrogen
            + doc_energy_per_ask_long_range_hydrogen
            + doc_carbon_tax_per_ask_long_range_hydrogen
        )

        # Average per category
        doc_total_per_ask_short_range_mean = (
            doc_non_energy_per_ask_short_range_mean
            + doc_energy_per_ask_short_range_mean
            + doc_carbon_tax_per_ask_short_range_mean
        )

        doc_total_per_ask_medium_range_mean = (
            doc_non_energy_per_ask_medium_range_mean
            + doc_energy_per_ask_medium_range_mean
            + doc_carbon_tax_per_ask_medium_range_mean
        )

        doc_total_per_ask_long_range_mean = (
            doc_non_energy_per_ask_long_range_mean
            + doc_energy_per_ask_long_range_mean
            + doc_carbon_tax_per_ask_long_range_mean
        )

        # total average

        doc_total_per_ask_mean = (
            doc_non_energy_per_ask_mean + doc_energy_per_ask_mean + doc_carbon_tax_per_ask_mean
        )

        self.df.loc[
            :, "doc_total_per_ask_short_range_dropin_fuel"
        ] = doc_total_per_ask_short_range_dropin_fuel
        self.df.loc[
            :, "doc_total_per_ask_medium_range_dropin_fuel"
        ] = doc_total_per_ask_medium_range_dropin_fuel
        self.df.loc[
            :, "doc_total_per_ask_long_range_dropin_fuel"
        ] = doc_total_per_ask_long_range_dropin_fuel
        self.df.loc[
            :, "doc_total_per_ask_short_range_hydrogen"
        ] = doc_total_per_ask_short_range_hydrogen
        self.df.loc[
            :, "doc_total_per_ask_medium_range_hydrogen"
        ] = doc_total_per_ask_medium_range_hydrogen
        self.df.loc[
            :, "doc_total_per_ask_long_range_hydrogen"
        ] = doc_total_per_ask_long_range_hydrogen
        self.df.loc[:, "doc_total_per_ask_short_range_mean"] = doc_total_per_ask_short_range_mean
        self.df.loc[:, "doc_total_per_ask_medium_range_mean"] = doc_total_per_ask_medium_range_mean
        self.df.loc[:, "doc_total_per_ask_long_range_mean"] = doc_total_per_ask_long_range_mean
        self.df.loc[:, "doc_total_per_ask_mean"] = doc_total_per_ask_mean

        return (
            doc_total_per_ask_short_range_dropin_fuel,
            doc_total_per_ask_medium_range_dropin_fuel,
            doc_total_per_ask_long_range_dropin_fuel,
            doc_total_per_ask_short_range_hydrogen,
            doc_total_per_ask_medium_range_hydrogen,
            doc_total_per_ask_long_range_hydrogen,
            doc_total_per_ask_short_range_mean,
            doc_total_per_ask_medium_range_mean,
            doc_total_per_ask_long_range_mean,
            doc_total_per_ask_mean,
        )


class DropInMeanMfsp(AeromapsModel):
    def __init__(self, name="dropin_mean_mfsp", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        biofuel_mean_mfsp: pd.Series = pd.Series(dtype="float64"),
        biofuel_mean_carbon_tax_per_l: pd.Series = pd.Series(dtype="float64"),
        biofuel_share: pd.Series = pd.Series(dtype="float64"),
        electrofuel_avg_cost_per_l: pd.Series = pd.Series(dtype="float64"),
        electrofuel_mfsp_carbon_tax_supplement: pd.Series = pd.Series(dtype="float64"),
        electrofuel_share: pd.Series = pd.Series(dtype="float64"),
        kerosene_market_price: pd.Series = pd.Series(dtype="float64"),
        kerosene_price_supplement_carbon_tax: pd.Series = pd.Series(dtype="float64"),
        kerosene_share: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series]:

        dropin_mean_mfsp = (
            biofuel_mean_mfsp.fillna(0) * biofuel_share / 100
            + electrofuel_avg_cost_per_l.fillna(0) * electrofuel_share / 100
            + kerosene_market_price.fillna(0) * kerosene_share / 100
        )

        dropin_mfsp_carbon_tax_supplement = (
            biofuel_mean_carbon_tax_per_l.fillna(0) * biofuel_share / 100
            + electrofuel_mfsp_carbon_tax_supplement.fillna(0) * electrofuel_share / 100
            + kerosene_price_supplement_carbon_tax.fillna(0) * kerosene_share / 100
        )

        self.df.loc[:, "dropin_mean_mfsp"] = dropin_mean_mfsp
        self.df.loc[:, "dropin_mfsp_carbon_tax_supplement"] = dropin_mfsp_carbon_tax_supplement

        return (dropin_mean_mfsp, dropin_mfsp_carbon_tax_supplement)



from typing import Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from aeromaps.models.base import AeromapsModel

# Vérifier l'appellation fuel ou drop-in fuel ou autre ?


class FuelDistribution(AeromapsModel):
    def __init__(self, name="aircraft_energy", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        biofuel_share_2020: float = 0.0,
        biofuel_share_2030: float = 0.0,
        biofuel_share_2040: float = 0.0,
        biofuel_share_2050: float = 0.0,
        electrofuel_share_2020: float = 0.0,
        electrofuel_share_2030: float = 0.0,
        electrofuel_share_2040: float = 0.0,
        electrofuel_share_2050: float = 0.0,
    ) -> Tuple[pd.Series, pd.Series, pd.Series]:
        """Fuel distribution calculation using interpolation functions"""

        reference_years = [2020, 2030, 2040, self.parameters.end_year]

        # Biofuel
        reference_values_biofuel = [
            biofuel_share_2020,
            biofuel_share_2030,
            biofuel_share_2040,
            biofuel_share_2050,
        ]
        biofuel_share_function = interp1d(
            reference_years, reference_values_biofuel, kind="quadratic"
        )
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "biofuel_share"] = 0
        for k in range(self.prospection_start_year, self.end_year + 1):
            if biofuel_share_function(k) <= 0:
                self.df.loc[k, "biofuel_share"] = 0
            else:
                self.df.loc[k, "biofuel_share"] = biofuel_share_function(k)

        # Electrofuel
        reference_values_electrofuel = [
            electrofuel_share_2020,
            electrofuel_share_2030,
            electrofuel_share_2040,
            electrofuel_share_2050,
        ]
        electrofuel_share_function = interp1d(
            reference_years, reference_values_electrofuel, kind="quadratic"
        )
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "electrofuel_share"] = 0
        for k in range(self.prospection_start_year, self.end_year + 1):
            if electrofuel_share_function(k) <= 0:
                self.df.loc[k, "electrofuel_share"] = 0
            else:
                self.df.loc[k, "electrofuel_share"] = electrofuel_share_function(k)

        biofuel_share = self.df["biofuel_share"]
        electrofuel_share = self.df["electrofuel_share"]

        # Kerosene
        kerosene_share = 100 - biofuel_share - electrofuel_share
        self.df.loc[:, "kerosene_share"] = kerosene_share

        return biofuel_share, electrofuel_share, kerosene_share



from dataclasses import dataclass
import os.path as pth
import pandas as pd
from aeromaps.utils.functions import _dict_from_json

filename = pth.join(
    pth.dirname(pth.abspath(__file__)), "..", "..", "resources", "data", "parameters.json"
)

data = _dict_from_json(file_name=filename)


@dataclass
class AirTransportParameters(object):

    # DataFrame data air on air traffic
    rpk_init: pd.Series = data["rpk_init"]
    ask_init: pd.Series = data["ask_init"]
    rtk_init: pd.Series = data["rtk_init"]
    pax_init: pd.Series = data["pax_init"]  # Not used
    freight_init: pd.Series = data["freight_init"]  # Not used
    energy_consumption_init: pd.Series = data["energy_consumption_init"]
    total_aircraft_distance_init: pd.Series = data["total_aircraft_distance_init"]

    # Other coefficients
    commercial_aviation_coefficient: float = 0.88

    # Load factor
    load_factor_end_year: float = 85.0  # [%]

    # Simple gains for fleet efficiency
    energy_per_ask_short_range_dropin_fuel_gain: float = 2  # [%]
    energy_per_ask_medium_range_dropin_fuel_gain: float = 2  # [%]
    energy_per_ask_long_range_dropin_fuel_gain: float = 2  # [%]

    # NOx and soot emission index for simple models [%] - Positive means augmentation
    emission_index_nox_dropin_fuel_evolution: float = 0.0  # [%]
    emission_index_nox_hydrogen_evolution: float = 0.0  # [%]
    emission_index_soot_dropin_fuel_evolution: float = 0.0  # [%]

    # Energy distribution [%]
    short_range_energy_share_2019: float = 30.6 * 0.85
    medium_range_energy_share_2019: float = 30.8 * 0.85
    long_range_energy_share_2019: float = 38.6 * 0.85
    freight_energy_share_2019: float = 15

    # RPK distribution [%]
    short_range_rpk_share_2019: float = 27.2
    medium_range_rpk_share_2019: float = 35.1
    long_range_rpk_share_2019: float = 37.7
    short_range_basicturbofan_share_2019: float = 88.4
    short_range_regionalturboprop_share_2019: float = 2.5
    short_range_regionalturbofan_share_2019: float = 9.1  # Model ?

    # Covid-19
    covid_start_year: int = 2020
    covid_rpk_drop_start_year: float = 66.0  # [%]
    covid_end_year: int = 2024
    covid_end_year_reference_rpk_ratio: float = 100.0  # [%]
    covid_load_factor_2020: float = 65.2  # [%]
    covid_energy_intensity_per_ask_increase_2020: float = 30.5  # [%]

    # Growth rate [%]
    growth_rate_2020_2030_short_range: float = 3.0
    growth_rate_2030_2040_short_range: float = 3.0
    growth_rate_2040_2050_short_range: float = 3.0
    growth_rate_2020_2030_medium_range: float = 3.0
    growth_rate_2030_2040_medium_range: float = 3.0
    growth_rate_2040_2050_medium_range: float = 3.0
    growth_rate_2020_2030_long_range: float = 3.0
    growth_rate_2030_2040_long_range: float = 3.0
    growth_rate_2040_2050_long_range: float = 3.0
    growth_rate_2020_2030_freight: float = 3.0
    growth_rate_2030_2040_freight: float = 3.0
    growth_rate_2040_2050_freight: float = 3.0

    # Short range distribution [%] - Not used
    short_range_basicturbofan_share_2030: float = 88.4
    short_range_basicturbofan_share_2040: float = 88.4
    short_range_basicturbofan_share_2050: float = 88.4
    short_range_regionalturboprop_share_2030: float = 2.5
    short_range_regionalturboprop_share_2040: float = 2.5
    short_range_regionalturboprop_share_2050: float = 2.5
    # Calcul pour ces valeurs
    # short_range_regionalturbofan_share_2030: float = 9.1
    # short_range_regionalturbofan_share_2040: float = 9.1
    # short_range_regionalturbofan_share_2050: float = 9.1

    # References
    growth_rate_2020_2030_reference: float = 3.0
    growth_rate_2030_2040_reference: float = 3.0
    growth_rate_2040_2050_reference: float = 3.0

    # RPK measures
    rpk_short_range_measures_final_impact: float = 0.0
    rpk_medium_range_measures_final_impact: float = 0.0
    rpk_long_range_measures_final_impact: float = 0.0
    rpk_short_range_measures_start_year: float = 2051
    rpk_medium_range_measures_start_year: float = 2051
    rpk_long_range_measures_start_year: float = 2051
    rpk_short_range_measures_duration: float = 5.0
    rpk_medium_range_measures_duration: float = 5.0
    rpk_long_range_measures_duration: float = 5.0

    # Operations
    operations_final_gain: float = 7.0  # [%]
    operations_start_year: float = 2020
    operations_duration: float = 40.0

    # Operations contrails
    operations_contrails_final_gain: float = 59.4  # [%]
    operations_contrails_final_overconsumption: float = 0.014  # [%]
    operations_contrails_start_year: float = 2051
    operations_contrails_duration: float = 15.0

    # Efficiency for hydrogen simple models
    hydrogen_final_market_share_short_range: float = 20.0  # [%]
    hydrogen_introduction_year_short_range: float = 2035
    hydrogen_final_market_share_medium_range: float = 20.0  # [%]
    hydrogen_introduction_year_medium_range: float = 2035
    hydrogen_final_market_share_long_range: float = 0.0  # [%]
    hydrogen_introduction_year_long_range: float = 2035
    relative_energy_per_ask_hydrogen_wrt_dropin_short_range: float = 1.0
    relative_energy_per_ask_hydrogen_wrt_dropin_medium_range: float = 1.0
    relative_energy_per_ask_hydrogen_wrt_dropin_long_range: float = 1.0
    fleet_renewal_duration: float = 25.0

    # Fuel distribution
    biofuel_share_2020: float = 0.0
    biofuel_share_2030: float = 0.0
    biofuel_share_2040: float = 0.0
    biofuel_share_2050: float = 0.0
    electrofuel_share_2020: float = 0.0
    electrofuel_share_2030: float = 0.0
    electrofuel_share_2040: float = 0.0
    electrofuel_share_2050: float = 0.0
    # Calculation for the following values
    # kerosene_share_2020: float = 100.0
    # kerosene_share_2030: float = 100.0
    # kerosene_share_2040: float = 100.0
    # kerosene_share_2050: float = 100.0

    # Energy efficiencies - Biofuel
    biofuel_ft_efficiency_2020: float = 0.46
    biofuel_ft_efficiency_2050: float = 0.46
    biofuel_atj_efficiency_2020: float = 0.48
    biofuel_atj_efficiency_2050: float = 0.48
    biofuel_hefa_oil_efficiency_2020: float = 0.66
    biofuel_hefa_oil_efficiency_2050: float = 0.66
    biofuel_hefa_fuel_efficiency_2020: float = 0.88
    biofuel_hefa_fuel_efficiency_2050: float = 0.88

    # Has to be integrated
    hydrogen_oil_hefa: float = 0.09

    # Energy selectivity - Biofuel
    # To be done with electrofuel
    biofuel_ft_selectivity: float = 0.5
    biofuel_atj_selectivity: float = 0.75
    biofuel_hefa_selectivity: float = 0.6

    # Energy production choices - Biofuel
    biofuel_hefa_fog_share_2020: float = 100.0
    biofuel_hefa_fog_share_2030: float = 100.0
    biofuel_hefa_fog_share_2040: float = 100.0
    biofuel_hefa_fog_share_2050: float = 100.0
    biofuel_hefa_others_share_2020: float = 0.0
    biofuel_hefa_others_share_2030: float = 0.0
    biofuel_hefa_others_share_2040: float = 0.0
    biofuel_hefa_others_share_2050: float = 0.0
    biofuel_ft_others_share_2020: float = 0.0
    biofuel_ft_others_share_2030: float = 0.0
    biofuel_ft_others_share_2040: float = 0.0
    biofuel_ft_others_share_2050: float = 0.0
    biofuel_ft_msw_share_2020: float = 0.0
    biofuel_ft_msw_share_2030: float = 0.0
    biofuel_ft_msw_share_2040: float = 0.0
    biofuel_ft_msw_share_2050: float = 0.0
    # Calculation for the following values
    # biofuel_atj_share_2020: float = 0.0
    # biofuel_atj_share_2030: float = 0.0
    # biofuel_atj_share_2040: float = 0.0
    # biofuel_atj_share_2050: float = 0.0

    # Energy efficiencies - Hydrogen
    electrolysis_efficiency_2020: float = 0.59
    electrolysis_efficiency_2050: float = 0.59
    liquefaction_efficiency_2020: float = 0.8
    liquefaction_efficiency_2050: float = 0.8

    # Energy production choices - Hydrogen
    hydrogen_electrolysis_share_2020: float = 2.0
    hydrogen_electrolysis_share_2030: float = 2.0
    hydrogen_electrolysis_share_2040: float = 2.0
    hydrogen_electrolysis_share_2050: float = 2.0
    hydrogen_gas_ccs_share_2020: float = 0.0
    hydrogen_gas_ccs_share_2030: float = 0.0
    hydrogen_gas_ccs_share_2040: float = 0.0
    hydrogen_gas_ccs_share_2050: float = 0.0
    hydrogen_coal_ccs_share_2020: float = 0.0
    hydrogen_coal_ccs_share_2030: float = 0.0
    hydrogen_coal_ccs_share_2040: float = 0.0
    hydrogen_coal_ccs_share_2050: float = 0.0
    hydrogen_gas_share_2020: float = 71.0
    hydrogen_gas_share_2030: float = 71.0
    hydrogen_gas_share_2040: float = 71.0
    hydrogen_gas_share_2050: float = 71.0
    # Calculation for the following values
    # hydrogen_coal_share_2020: float = 27.0
    # hydrogen_coal_share_2030: float = 27.0
    # hydrogen_coal_share_2040: float = 27.0
    # hydrogen_coal_share_2050: float = 27.0

    # Energy efficiencies - Electrofuel
    electrofuel_hydrogen_efficiency_2020: float = 0.67
    electrofuel_hydrogen_efficiency_2050: float = 0.67

    # Energy emissions - Biofuel
    biofuel_hefa_fog_emission_factor_2020: float = 20.7
    biofuel_hefa_fog_emission_factor_2050: float = 20.7
    biofuel_hefa_others_emission_factor_2020: float = 61
    biofuel_hefa_others_emission_factor_2050: float = 61
    biofuel_ft_others_emission_factor_2020: float = 7.7
    biofuel_ft_others_emission_factor_2050: float = 7.7
    biofuel_ft_msw_emission_factor_2020: float = 27.6
    biofuel_ft_msw_emission_factor_2050: float = 27.6
    biofuel_atj_emission_factor_2020: float = 52.2
    biofuel_atj_emission_factor_2050: float = 52.2

    # Energy emissions - Electricity [gCO2/kWh]
    electricity_emission_factor_2020: float = 429
    electricity_emission_factor_2030: float = 200
    electricity_emission_factor_2040: float = 200
    electricity_emission_factor_2050: float = 200

    # Energy emissions - Hydrogen [gCO2/MJ]
    hydrogen_gas_ccs_emission_factor: float = 3.7 / 120 * 1000
    hydrogen_coal_ccs_emission_factor: float = 4.9 / 120 * 1000
    hydrogen_gas_emission_factor: float = 12.0 / 120 * 1000
    hydrogen_coal_emission_factor: float = 23.0 / 120 * 1000



from typing import Tuple

import pandas as pd


from aeromaps.models.base import AeromapsModel


class TotalAircraftDistance(AeromapsModel):
    def __init__(self, name="total_aircraft_distance", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        rpk: pd.Series = pd.Series(dtype="float64"),
        total_aircraft_distance_init: pd.Series = pd.Series(dtype="float64"),
    ) -> pd.Series:
        """Total aircraft distance calculation."""

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "total_aircraft_distance"] = total_aircraft_distance_init.loc[k]

        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "total_aircraft_distance"] = (
                rpk.loc[k]
                / rpk.loc[self.prospection_start_year - 1]
                * self.df.loc[self.prospection_start_year - 1, "total_aircraft_distance"]
            )

        total_aircraft_distance = self.df["total_aircraft_distance"]

        return total_aircraft_distance



from typing import Tuple

import numpy as np
import pandas as pd

from aeromaps.models.base import AeromapsModel


class ASK(AeromapsModel):
    def __init__(self, name="ask", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        load_factor: pd.Series = pd.Series(dtype="float64"),
        rpk: pd.Series = pd.Series(dtype="float64"),
        rpk_short_range: pd.Series = pd.Series(dtype="float64"),
        rpk_medium_range: pd.Series = pd.Series(dtype="float64"),
        rpk_long_range: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """ASK calculation."""

        # ASK calculation
        ask = rpk / (load_factor / 100)
        self.df.loc[:, "ask"] = ask

        # Short range
        for k in range(self.historic_start_year, self.end_year + 1):
            self.df.loc[k, "ask_short_range"] = rpk_short_range.loc[k] / (load_factor.loc[k] / 100)

        # Medium range
        for k in range(self.historic_start_year, self.end_year + 1):
            self.df.loc[k, "ask_medium_range"] = rpk_medium_range.loc[k] / (
                load_factor.loc[k] / 100
            )

        # Long range
        for k in range(self.historic_start_year, self.end_year + 1):
            self.df.loc[k, "ask_long_range"] = rpk_long_range.loc[k] / (load_factor.loc[k] / 100)

        ask_short_range = self.df["ask_short_range"]
        ask_medium_range = self.df["ask_medium_range"]
        ask_long_range = self.df["ask_long_range"]

        return ask, ask_short_range, ask_medium_range, ask_long_range



from typing import Tuple

import numpy as np
import pandas as pd

from aeromaps.models.base import AeromapsModel


class BiomassConsumption(AeromapsModel):
    def __init__(self, name="biomass_consumption", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_biofuel: pd.Series = pd.Series(dtype="float64"),
        biofuel_hefa_fog_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_hefa_others_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_ft_others_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_ft_msw_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_atj_share: pd.Series = pd.Series(dtype="float64"),
        biofuel_ft_efficiency: pd.Series = pd.Series(dtype="float64"),
        biofuel_atj_efficiency: pd.Series = pd.Series(dtype="float64"),
        biofuel_hefa_oil_efficiency: pd.Series = pd.Series(dtype="float64"),
        biofuel_hefa_fuel_efficiency: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, float]:
        """Biomass consumption calculation."""

        biomass_ft_consumption = (
            (biofuel_ft_others_share + biofuel_ft_msw_share)
            / 100
            * energy_consumption_biofuel
            / 10**12
            / biofuel_ft_efficiency
        )
        biomass_atj_consumption = (
            biofuel_atj_share / 100 * energy_consumption_biofuel / 10**12 / biofuel_atj_efficiency
        )
        biomass_hefa_fog_consumption = (
            biofuel_hefa_fog_share
            / 100
            * energy_consumption_biofuel
            / 10**12
            / biofuel_hefa_fuel_efficiency
        )
        biomass_hefa_others_consumption = (
            biofuel_hefa_others_share
            / 100
            * energy_consumption_biofuel
            / 10**12
            / biofuel_hefa_fuel_efficiency
            / biofuel_hefa_oil_efficiency
        )
        biomass_consumption = (
            biomass_ft_consumption
            + biomass_atj_consumption
            + biomass_hefa_fog_consumption
            + biomass_hefa_others_consumption
        )

        self.df.loc[:, "biomass_ft_consumption"] = biomass_ft_consumption
        self.df.loc[:, "biomass_atj_consumption"] = biomass_atj_consumption
        self.df.loc[:, "biomass_hefa_fog_consumption"] = biomass_hefa_fog_consumption
        self.df.loc[:, "biomass_hefa_others_consumption"] = biomass_hefa_others_consumption
        self.df.loc[:, "biomass_consumption"] = biomass_consumption

        biomass_consumption_end_year = self.df.loc[self.end_year, "biomass_consumption"]

        self.float_outputs["biomass_consumption_end_year"] = biomass_consumption_end_year

        return (
            biomass_ft_consumption,
            biomass_atj_consumption,
            biomass_hefa_fog_consumption,
            biomass_hefa_others_consumption,
            biomass_consumption,
            biomass_consumption_end_year,
        )


class ElectricityConsumption(AeromapsModel):
    def __init__(self, name="electricity_consumption", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        hydrogen_electrolysis_share: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_hydrogen: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_electrofuel: pd.Series = pd.Series(dtype="float64"),
        electrolysis_efficiency: float = 0.0,
        liquefaction_efficiency: float = 0.0,
        electrofuel_hydrogen_efficiency: float = 0.0,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, float]:
        """Electricity consumption calculation."""

        electricity_hydrogen_consumption = (
            hydrogen_electrolysis_share
            / 100
            * energy_consumption_hydrogen
            / 10**12
            / electrolysis_efficiency
            / liquefaction_efficiency
        )
        electricity_electrofuel_consumption = (
            energy_consumption_electrofuel
            / 10**12
            / electrolysis_efficiency
            / electrofuel_hydrogen_efficiency
        )
        electricity_consumption = (
            electricity_hydrogen_consumption + electricity_electrofuel_consumption
        )

        self.df.loc[:, "electricity_hydrogen_consumption"] = electricity_hydrogen_consumption
        self.df.loc[:, "electricity_electrofuel_consumption"] = electricity_electrofuel_consumption
        self.df.loc[:, "electricity_consumption"] = electricity_consumption

        electricity_consumption_end_year = self.df.loc[self.end_year, "electricity_consumption"]

        self.float_outputs["electricity_consumption_end_year"] = electricity_consumption_end_year

        return (
            electricity_hydrogen_consumption,
            electricity_electrofuel_consumption,
            electricity_consumption,
            electricity_consumption_end_year,
        )



import matplotlib.pyplot as plt
import numpy as np
from .constants import plot_1_x, plot_1_y


class AirTransportCO2EmissionsPlot:
    def __init__(self, data):
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.fig, self.ax = plt.subplots(
            figsize=(plot_1_x, plot_1_y),
        )
        self.create_plot()

    def create_plot(self):
        (self.line_co2_emissions_including_sobriety,) = self.ax.plot(
            self.years,
            self.df["co2_emissions_2019technology"],
            color="black",
            linestyle="--",
            linewidth=1,
        )

        (self.line_co2_emissions_including_technology,) = self.ax.plot(
            self.years,
            self.df["co2_emissions_including_aircraft_efficiency"],
            color="black",
            linestyle="--",
            linewidth=1,
        )

        (self.line_co2_emissions_including_load_factor,) = self.ax.plot(
            self.years,
            self.df["co2_emissions_including_load_factor"],
            color="black",
            linestyle="--",
            linewidth=1,
        )

        (self.line_co2_emissions_including_energy,) = self.ax.plot(
            self.years,
            self.df["co2_emissions_including_energy"],
            color="black",
            linestyle="--",
            linewidth=1,
        )

        self.ax.plot(
            self.historic_years,
            self.df.loc[self.historic_years, "co2_emissions"],
            color="black",
            linestyle="-",
            label="Historical emissions",
            linewidth=3,
            zorder=4,
        )

        (self.line_co2_emissions_no_action,) = self.ax.plot(
            self.prospective_years,
            self.df.loc[self.prospective_years, "co2_emissions_2019technology_baseline3"],
            color="red",
            linestyle="-",
            label="Emissions at 2019 technological level with trend air traffic growth",
            linewidth=3,
            zorder=3,
        )

        (self.line_co2_emissions,) = self.ax.plot(
            self.prospective_years,
            self.df.loc[self.prospective_years, "co2_emissions"],
            color="green",
            linestyle="-",
            label="Projected emissions including all levers of action",
            linewidth=3,
            zorder=3,
        )

        # Fill between

        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_2019technology_baseline3"],
            self.df["co2_emissions_2019technology"],
            color="lightskyblue",
            label="Demand/supply side management",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_2019technology"],
            self.df["co2_emissions_including_aircraft_efficiency"],
            color="gold",
            label="Aircraft efficiency",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_including_aircraft_efficiency"],
            self.df["co2_emissions_including_load_factor"],
            color="orange",
            label="Fleet operations and load factor",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_including_load_factor"],
            self.df["co2_emissions"],
            color="yellowgreen",
            label="Aircraft energy",
        )

        self.ax.grid()
        self.ax.set_title("Evolution of annual CO2 emissions from air transport")
        self.ax.set_xlabel("Year")
        self.ax.set_ylabel("Annual CO2 emissions [MtCO2]")
        ax = plt.gca()
        self.ax.legend(loc=2)
        self.ax.set_xlim(self.years[0], self.years[-1])

        self.fig.canvas.header_visible = False
        self.fig.canvas.toolbar_position = "bottom"
        # self.fig.canvas.layout.width = "auto"
        # self.fig.canvas.layout.height = "auto"
        self.fig.tight_layout()

    def update(self, data):
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.line_co2_emissions_including_sobriety.set_ydata(
            self.df["co2_emissions_2019technology"]
        )

        self.line_co2_emissions_including_technology.set_ydata(
            self.df["co2_emissions_including_aircraft_efficiency"]
        )

        self.line_co2_emissions_including_load_factor.set_ydata(
            self.df["co2_emissions_including_load_factor"]
        )

        self.line_co2_emissions_including_energy.set_ydata(
            self.df["co2_emissions_including_energy"]
        )

        self.line_co2_emissions.set_ydata(self.df.loc[self.prospective_years, "co2_emissions"])

        self.ax.collections.clear()

        # Fill between
        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_2019technology_baseline3"],
            self.df["co2_emissions_2019technology"],
            color="lightskyblue",
            label="Demand/supply side management",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_2019technology"],
            self.df["co2_emissions_including_aircraft_efficiency"],
            color="gold",
            label="Aircraft efficiency",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_including_aircraft_efficiency"],
            self.df["co2_emissions_including_load_factor"],
            color="orange",
            label="Fleet operations and load factor",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_emissions_including_load_factor"],
            self.df["co2_emissions"],
            color="yellowgreen",
            label="Aircraft energy",
        )

        self.ax.relim()
        self.ax.autoscale_view()
        self.fig.canvas.draw()


class AirTransportClimateImpactsPlot:
    def __init__(self, data):
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.fig, self.ax = plt.subplots(
            figsize=(plot_1_x, plot_1_y),
        )
        self.create_plot()

    def create_plot(self):
        (self.line_co2_erf,) = self.ax.plot(
            self.years,
            self.df["co2_erf"],
            color="black",
            linestyle="--",
            linewidth=1,
        )

        (self.line_co2_h2o_erf,) = self.ax.plot(
            self.years,
            self.df["co2_h2o_erf"],
            color="black",
            linestyle="--",
            linewidth=1,
        )

        (self.line_co2_h2o_nox_erf,) = self.ax.plot(
            self.years,
            self.df["co2_h2o_nox_erf"],
            color="black",
            linestyle="--",
            linewidth=1,
        )

        (self.line_co2_h2o_nox_contrails_erf,) = self.ax.plot(
            self.years,
            self.df["co2_h2o_nox_contrails_erf"],
            linestyle="None",
        )

        (self.line_aerosol_erf,) = self.ax.plot(
            self.years,
            self.df["aerosol_erf"],
            linestyle="None",
        )

        self.ax.plot(
            self.historic_years,
            self.df.loc[self.historic_years, "total_erf"],
            color="black",
            linestyle="-",
            label="Net ERF - History",
            linewidth=4,
        )

        (self.line_total_erf,) = self.ax.plot(
            self.prospective_years,
            self.df.loc[self.prospective_years, "total_erf"],
            color="green",
            linestyle="-",
            label="Net ERF - Projections",
            linewidth=4,
        )

        # Fill between
        self.ax.fill_between(
            self.years,
            np.zeros(len(self.years)),
            self.df["co2_erf"],
            color="tomato",
            label="CO2",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_erf"],
            self.df["co2_h2o_erf"],
            color="lightskyblue",
            label="H2O",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_h2o_erf"],
            self.df["co2_h2o_nox_erf"],
            color="yellowgreen",
            label="NOx",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_h2o_nox_erf"],
            self.df["co2_h2o_nox_contrails_erf"],
            color="gold",
            label="Contrails",
        )

        self.ax.fill_between(
            self.years,
            np.zeros(len(self.years)),
            self.df["aerosol_erf"],
            color="darkblue",
            label="Aerosols",
        )

        self.ax.grid()
        self.ax.set_title(
            "Evolution of climate impacts (via effective radiative forcing) from air transport"
        )
        self.ax.set_xlabel("Year")
        self.ax.set_ylabel("Effective radiative forcing [mW/m²]")
        self.ax.legend(loc=2)
        self.ax.set_xlim(self.years[0], self.years[-1])

        self.fig.canvas.header_visible = False
        self.fig.canvas.toolbar_position = "bottom"
        # self.fig.canvas.layout.width = "auto"
        # self.fig.canvas.layout.height = "auto"
        self.fig.tight_layout()

    def update(self, data):
        self.df = data["vector_outputs"]
        self.float_outputs = data["float_outputs"]
        self.years = data["years"]["full_years"]
        self.historic_years = data["years"]["historic_years"]
        self.prospective_years = data["years"]["prospective_years"]

        self.line_co2_erf.set_ydata(self.df["co2_erf"])

        self.line_co2_h2o_erf.set_ydata(self.df["co2_h2o_erf"])

        self.line_co2_h2o_nox_erf.set_ydata(self.df["co2_h2o_nox_erf"])

        self.line_co2_h2o_nox_contrails_erf.set_ydata(self.df["co2_h2o_nox_contrails_erf"])

        self.line_aerosol_erf.set_ydata(self.df["aerosol_erf"])

        self.line_total_erf.set_ydata(self.df.loc[self.prospective_years, "total_erf"])

        self.ax.collections.clear()

        # Fill between
        self.ax.fill_between(
            self.years,
            np.zeros(len(self.years)),
            self.df["co2_erf"],
            color="tomato",
            label="CO2",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_erf"],
            self.df["co2_h2o_erf"],
            color="lightskyblue",
            label="H2O",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_h2o_erf"],
            self.df["co2_h2o_nox_erf"],
            color="yellowgreen",
            label="NOx",
        )

        self.ax.fill_between(
            self.years,
            self.df["co2_h2o_nox_erf"],
            self.df["co2_h2o_nox_contrails_erf"],
            color="gold",
            label="Contrails",
        )

        self.ax.fill_between(
            self.years,
            np.zeros(len(self.years)),
            self.df["aerosol_erf"],
            color="darkblue",
            label="Aerosols",
        )

        self.ax.relim()
        self.ax.autoscale_view()
        self.fig.canvas.draw()



from gemseo.core.discipline import MDODiscipline

from aeromaps.models.parameters import AllParameters
from aeromaps.models.sustainability_assessment.climate.carbon_budgets import GrossCarbonBudget
from aeromaps.models.sustainability_assessment.climate.equivalent_carbon_budgets import (
    EquivalentGrossCarbonBudget,
)
from aeromaps.models.sustainability_assessment.energy.resources_availability import (
    BiomassAvailability,
    ElectricityAvailability,
)
from aeromaps.core.gemseo import AeromapsModelWrapper
from gemseo.api import create_discipline, generate_n2_plot, create_mda

disc1 = AeromapsModelWrapper(model=GrossCarbonBudget(parameters=AllParameters()))
disc2 = AeromapsModelWrapper(model=EquivalentGrossCarbonBudget(parameters=AllParameters()))
disc3 = AeromapsModelWrapper(model=BiomassAvailability(parameters=AllParameters()))
disc4 = AeromapsModelWrapper(model=ElectricityAvailability(parameters=AllParameters()))

disciplines = [disc1, disc2, disc3, disc4]

# generate_n2_plot(disciplines)

# process = create_discipline(
#     "MDOChain", disciplines=disciplines, grammar_type=MDODiscipline.SIMPLE_GRAMMAR_TYPE
# )
process = create_mda(
    "MDAChain", disciplines=disciplines, grammar_type=MDODiscipline.SIMPLE_GRAMMAR_TYPE
)

outputs = process.execute()

print(outputs)

disc1.model.parameters.net_carbon_budget = 300.0

for disc in disciplines:
    disc.update_defaults()

# process._update_default_inputs()
process._set_default_inputs()

outputs = process.execute()

print(outputs)



import numpy as np
from dataclasses import dataclass

import pandas as pd
import ipydatagrid as dg
from ipytree import Tree, Node
import matplotlib.pyplot as plt

import ipywidgets as ipw

from aeromaps.models.base import AeromapsModel

AIRCRAFT_COLUMNS = [
    "Name",
    "EIS Year",
    "Consumption evolution [%]",
    "NOx evolution [%]",
    "Soot evolution [%]",
    "Non-Energy DOC evolution [%]",
    "Cruise altitude [m]",
    "Energy Type",
]
SUBCATEGORY_COLUMNS = ["Name", "Share [%]"]


@dataclass
class AircraftParameters:
    entry_into_service_year: float = None
    consumption_evolution: float = None
    nox_evolution: float = None
    soot_evolution: float = None
    doc_non_energy_evolution: float = None
    cruise_altitude: float = None


@dataclass
class ReferenceAircraftParameters:
    energy_per_ask: float = None
    emission_index_nox: float = None
    emission_index_soot: float = None
    doc_non_energy_base: float = None
    entry_into_service_year: float = None
    cruise_altitude: float = None


@dataclass
class SubcategoryParameters:
    share: float = None


@dataclass
class CategoryParameters:
    life: float
    limit: float = 2


class FleetModel(AeromapsModel):
    def __init__(self, name="fleet_model", fleet=None, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.fleet = fleet

    def compute(
        self,
    ):
        # Start from empty dataframe
        self.df = self.df.filter([])

        # Compute single aircraft shares
        self._compute_single_aircraft_share()

        # Compute aircraft shares
        self._compute_aircraft_share()

        # Compute energy consumption and share per subcategory with respect to energy type
        self._compute_energy_consumption_and_share_wrt_energy_type()

        # Compute non energy direct operating costs (DOC) and share per subcategory with respect to energy type
        self._compute_doc_non_energy()

        # Compute mean energy consumption per category with respect to energy type
        self._compute_mean_energy_consumption_per_category_wrt_energy_type()

        # Compute mean non energy direct operating cost (DOC) per category with respect to energy type
        self._compute_mean_doc_non_energy()

    def _compute_energy_consumption_and_share_wrt_energy_type(self):
        # Energy consumption calculations for drop-in fuel and hydrogen
        for category in self.fleet.categories.values():
            # Reference aircraft information
            ref_old_aircraft_share = self.df[
                category.name
                + ":"
                + category.subcategories[0].name
                + ":"
                + "old_reference:aircraft_share"
            ]
            ref_recent_aircraft_share = self.df[
                category.name
                + ":"
                + category.subcategories[0].name
                + ":"
                + "recent_reference:aircraft_share"
            ]

            recent_reference_aircraft_energy_consumption = category.subcategories[
                0
            ].recent_reference_aircraft.energy_per_ask

            for i, subcategory in category.subcategories.items():
                # Initialization
                if i == 0:
                    self.df[category.name + ":" + subcategory.name + ":energy_consumption"] = (
                        subcategory.old_reference_aircraft.energy_per_ask
                        * ref_old_aircraft_share
                        / 100
                        + subcategory.recent_reference_aircraft.energy_per_ask
                        * ref_recent_aircraft_share
                        / 100
                    )
                    # Initial energy consumption
                    self.df[
                        category.name + ":" + subcategory.name + ":energy_consumption:dropin_fuel"
                    ] = self.df[category.name + ":" + subcategory.name + ":energy_consumption"]
                    self.df[
                        category.name + ":" + subcategory.name + ":energy_consumption:hydrogen"
                    ] = 0.0
                    # Initial shares
                    self.df[category.name + ":" + subcategory.name + ":share:total"] = (
                        ref_old_aircraft_share + ref_recent_aircraft_share
                    )
                    self.df[category.name + ":" + subcategory.name + ":share:dropin_fuel"] = (
                        ref_old_aircraft_share + ref_recent_aircraft_share
                    )
                    self.df[category.name + ":" + subcategory.name + ":share:hydrogen"] = 0.0

                else:
                    self.df[category.name + ":" + subcategory.name + ":energy_consumption"] = 0.0
                    # Initial energy consumption
                    self.df[
                        category.name + ":" + subcategory.name + ":energy_consumption:dropin_fuel"
                    ] = self.df[category.name + ":" + subcategory.name + ":energy_consumption"]
                    self.df[
                        category.name + ":" + subcategory.name + ":energy_consumption:hydrogen"
                    ] = self.df[category.name + ":" + subcategory.name + ":energy_consumption"]
                    # Initial shares
                    self.df[category.name + ":" + subcategory.name + ":share:total"] = 0.0
                    self.df[category.name + ":" + subcategory.name + ":share:dropin_fuel"] = 0.0
                    self.df[category.name + ":" + subcategory.name + ":share:hydrogen"] = 0.0

                for aircraft in subcategory.aircraft.values():

                    for k in self.df.index:

                        if (
                            self.df.loc[
                                k,
                                category.name
                                + ":"
                                + subcategory.name
                                + ":"
                                + aircraft.name
                                + ":aircraft_share",
                            ]
                            != 0.0
                        ):

                            self.df.loc[
                                k, category.name + ":" + subcategory.name + ":share:total"
                            ] += self.df.loc[
                                k,
                                category.name
                                + ":"
                                + subcategory.name
                                + ":"
                                + aircraft.name
                                + ":aircraft_share",
                            ]

                            self.df.loc[
                                k, category.name + ":" + subcategory.name + ":energy_consumption"
                            ] += (
                                recent_reference_aircraft_energy_consumption
                                * (1 + float(aircraft.parameters.consumption_evolution) / 100)
                                * self.df.loc[
                                    k,
                                    category.name
                                    + ":"
                                    + subcategory.name
                                    + ":"
                                    + aircraft.name
                                    + ":aircraft_share",
                                ]
                                / 100
                            )

                            if aircraft.energy_type == "DROP_IN_FUEL":
                                self.df.loc[
                                    k, category.name + ":" + subcategory.name + ":share:dropin_fuel"
                                ] += (
                                    self.df.loc[
                                        k,
                                        category.name
                                        + ":"
                                        + subcategory.name
                                        + ":"
                                        + aircraft.name
                                        + ":aircraft_share",
                                    ]
                                    # / self.df.loc[
                                    #     k, category.name + ":" + subcategory.name + ":share:total"
                                    # ]
                                    # * 100
                                )
                                self.df.loc[
                                    k,
                                    category.name
                                    + ":"
                                    + subcategory.name
                                    + ":energy_consumption:dropin_fuel",
                                ] += (
                                    recent_reference_aircraft_energy_consumption
                                    * (1 + float(aircraft.parameters.consumption_evolution) / 100)
                                    * self.df.loc[
                                        k,
                                        category.name
                                        + ":"
                                        + subcategory.name
                                        + ":"
                                        + aircraft.name
                                        + ":aircraft_share",
                                    ]
                                    / 100
                                )

                            if aircraft.energy_type == "HYDROGEN":
                                self.df.loc[
                                    k, category.name + ":" + subcategory.name + ":share:hydrogen"
                                ] += (
                                    self.df.loc[
                                        k,
                                        category.name
                                        + ":"
                                        + subcategory.name
                                        + ":"
                                        + aircraft.name
                                        + ":aircraft_share",
                                    ]
                                    # / self.df.loc[
                                    #     k, category.name + ":" + subcategory.name + ":share:total"
                                    # ]
                                    # * 100
                                )
                                self.df.loc[
                                    k,
                                    category.name
                                    + ":"
                                    + subcategory.name
                                    + ":energy_consumption:hydrogen",
                                ] += (
                                    recent_reference_aircraft_energy_consumption
                                    * (1 + float(aircraft.parameters.consumption_evolution) / 100)
                                    * self.df.loc[
                                        k,
                                        category.name
                                        + ":"
                                        + subcategory.name
                                        + ":"
                                        + aircraft.name
                                        + ":aircraft_share",
                                    ]
                                    / 100
                                )

                # Energy shares per category
                var_name = category.name + ":share:dropin_fuel"
                if var_name in self.df:
                    # Dropin
                    self.df[category.name + ":share:dropin_fuel"] += self.df[
                        category.name + ":" + subcategory.name + ":share:dropin_fuel"
                    ]
                else:
                    # Dropin
                    self.df[category.name + ":share:dropin_fuel"] = self.df[
                        category.name + ":" + subcategory.name + ":share:dropin_fuel"
                    ]

                var_name = category.name + ":share:hydrogen"
                if var_name in self.df:
                    # Hydrogen
                    self.df[category.name + ":share:hydrogen"] += self.df[
                        category.name + ":" + subcategory.name + ":share:hydrogen"
                    ]
                else:
                    # Hydrogen
                    self.df[category.name + ":share:hydrogen"] = self.df[
                        category.name + ":" + subcategory.name + ":share:hydrogen"
                    ]

    def _compute_doc_non_energy(self):
        # Non-energy DOC calculations for drop-in fuel and hydrogen
        for category in self.fleet.categories.values():
            # Reference aircraft information
            ref_old_aircraft_share = self.df[
                category.name
                + ":"
                + category.subcategories[0].name
                + ":"
                + "old_reference:aircraft_share"
            ]
            ref_recent_aircraft_share = self.df[
                category.name
                + ":"
                + category.subcategories[0].name
                + ":"
                + "recent_reference:aircraft_share"
            ]

            recent_reference_aircraft_doc_non_energy = category.subcategories[
                0
            ].recent_reference_aircraft.doc_non_energy_base

            for i, subcategory in category.subcategories.items():
                # Initialization
                if i == 0:
                    self.df[category.name + ":" + subcategory.name + ":doc_non_energy"] = (
                        subcategory.old_reference_aircraft.doc_non_energy_base
                        * ref_old_aircraft_share
                        / 100
                        + subcategory.recent_reference_aircraft.doc_non_energy_base
                        * ref_recent_aircraft_share
                        / 100
                    )
                    # Initial energy consumption
                    self.df[
                        category.name + ":" + subcategory.name + ":doc_non_energy:dropin_fuel"
                    ] = self.df[category.name + ":" + subcategory.name + ":doc_non_energy"]
                    self.df[
                        category.name + ":" + subcategory.name + ":doc_non_energy:hydrogen"
                    ] = 0.0
                    # Initial shares
                    # self.df[category.name + ":" + subcategory.name + ":share:total"] = (
                    #     ref_old_aircraft_share + ref_recent_aircraft_share
                    # )
                    # self.df[category.name + ":" + subcategory.name + ":share:dropin_fuel"] = (
                    #     ref_old_aircraft_share + ref_recent_aircraft_share
                    # )
                    # self.df[category.name + ":" + subcategory.name + ":share:hydrogen"] = 0.0

                else:
                    self.df[category.name + ":" + subcategory.name + ":doc_non_energy"] = 0.0
                    # Initial energy consumption
                    self.df[
                        category.name + ":" + subcategory.name + ":doc_non_energy:dropin_fuel"
                    ] = self.df[category.name + ":" + subcategory.name + ":doc_non_energy"]
                    self.df[
                        category.name + ":" + subcategory.name + ":doc_non_energy:hydrogen"
                    ] = self.df[category.name + ":" + subcategory.name + ":doc_non_energy"]
                    # Initial shares
                    # self.df[category.name + ":" + subcategory.name + ":share:total"] = 0.0
                    # self.df[category.name + ":" + subcategory.name + ":share:dropin_fuel"] = 0.0
                    # self.df[category.name + ":" + subcategory.name + ":share:hydrogen"] = 0.0

                for aircraft in subcategory.aircraft.values():

                    for k in self.df.index:

                        if (
                            self.df.loc[
                                k,
                                category.name
                                + ":"
                                + subcategory.name
                                + ":"
                                + aircraft.name
                                + ":aircraft_share",
                            ]
                            != 0.0
                        ):

                            # self.df.loc[
                            #     k, category.name + ":" + subcategory.name + ":share:total"
                            # ] += self.df.loc[
                            #     k,
                            #     category.name
                            #     + ":"
                            #     + subcategory.name
                            #     + ":"
                            #     + aircraft.name
                            #     + ":aircraft_share",
                            # ]

                            self.df.loc[
                                k, category.name + ":" + subcategory.name + ":doc_non_energy"
                            ] += (
                                recent_reference_aircraft_doc_non_energy
                                * (1 + float(aircraft.parameters.doc_non_energy_evolution) / 100)
                                * self.df.loc[
                                    k,
                                    category.name
                                    + ":"
                                    + subcategory.name
                                    + ":"
                                    + aircraft.name
                                    + ":aircraft_share",
                                ]
                                / 100
                            )

                            if aircraft.energy_type == "DROP_IN_FUEL":
                                # self.df.loc[
                                #     k, category.name + ":" + subcategory.name + ":share:dropin_fuel"
                                # ] += (
                                #     self.df.loc[
                                #         k,
                                #         category.name
                                #         + ":"
                                #         + subcategory.name
                                #         + ":"
                                #         + aircraft.name
                                #         + ":aircraft_share",
                                #     ]
                                #     # / self.df.loc[
                                #     #     k, category.name + ":" + subcategory.name + ":share:total"
                                #     # ]
                                #     # * 100
                                # )
                                self.df.loc[
                                    k,
                                    category.name
                                    + ":"
                                    + subcategory.name
                                    + ":doc_non_energy:dropin_fuel",
                                ] += (
                                    recent_reference_aircraft_doc_non_energy
                                    * (
                                        1
                                        + float(aircraft.parameters.doc_non_energy_evolution) / 100
                                    )
                                    * self.df.loc[
                                        k,
                                        category.name
                                        + ":"
                                        + subcategory.name
                                        + ":"
                                        + aircraft.name
                                        + ":aircraft_share",
                                    ]
                                    / 100
                                )

                            if aircraft.energy_type == "HYDROGEN":
                                # self.df.loc[
                                #     k, category.name + ":" + subcategory.name + ":share:hydrogen"
                                # ] += (
                                #     self.df.loc[
                                #         k,
                                #         category.name
                                #         + ":"
                                #         + subcategory.name
                                #         + ":"
                                #         + aircraft.name
                                #         + ":aircraft_share",
                                #     ]
                                #     # / self.df.loc[
                                #     #     k, category.name + ":" + subcategory.name + ":share:total"
                                #     # ]
                                #     # * 100
                                # )
                                self.df.loc[
                                    k,
                                    category.name
                                    + ":"
                                    + subcategory.name
                                    + ":doc_non_energy:hydrogen",
                                ] += (
                                    recent_reference_aircraft_doc_non_energy
                                    * (
                                        1
                                        + float(aircraft.parameters.doc_non_energy_evolution) / 100
                                    )
                                    * self.df.loc[
                                        k,
                                        category.name
                                        + ":"
                                        + subcategory.name
                                        + ":"
                                        + aircraft.name
                                        + ":aircraft_share",
                                    ]
                                    / 100
                                )

                # # Energy shares per category
                # var_name = category.name + ":share:dropin_fuel"
                # if var_name in self.df:
                #     # Dropin
                #     self.df[category.name + ":share:dropin_fuel"] += self.df[
                #         category.name + ":" + subcategory.name + ":share:dropin_fuel"
                #     ]
                # else:
                #     # Dropin
                #     self.df[category.name + ":share:dropin_fuel"] = self.df[
                #         category.name + ":" + subcategory.name + ":share:dropin_fuel"
                #     ]
                #
                # var_name = category.name + ":share:hydrogen"
                # if var_name in self.df:
                #     # Hydrogen
                #     self.df[category.name + ":share:hydrogen"] += self.df[
                #         category.name + ":" + subcategory.name + ":share:hydrogen"
                #     ]
                # else:
                #     # Hydrogen
                #     self.df[category.name + ":share:hydrogen"] = self.df[
                #         category.name + ":" + subcategory.name + ":share:hydrogen"
                #     ]

    def _compute_mean_energy_consumption_per_category_wrt_energy_type(self):
        for category in self.fleet.categories.values():
            # Mean energy consumption per category
            # Initialization
            self.df[category.name + ":energy_consumption:dropin_fuel"] = 0.0
            self.df[category.name + ":energy_consumption:hydrogen"] = 0.0
            # Calculation
            for subcategory in category.subcategories.values():
                # TODO: verify aircraft order
                for k in self.df.index:
                    if self.df.loc[k, category.name + ":share:dropin_fuel"] != 0.0:
                        self.df.loc[
                            k, category.name + ":energy_consumption:dropin_fuel"
                        ] += self.df.loc[
                            k,
                            category.name
                            + ":"
                            + subcategory.name
                            + ":energy_consumption:dropin_fuel",
                        ] / (
                            self.df.loc[k, category.name + ":share:dropin_fuel"] / 100
                        )
                    else:
                        self.df.loc[k, category.name + ":energy_consumption:dropin_fuel"] = 0.0

                    if self.df.loc[k, category.name + ":share:hydrogen"] != 0.0:
                        self.df.loc[
                            k, category.name + ":energy_consumption:hydrogen"
                        ] += self.df.loc[
                            k,
                            category.name + ":" + subcategory.name + ":energy_consumption:hydrogen",
                        ] / (
                            self.df.loc[k, category.name + ":share:hydrogen"] / 100
                        )
                    else:
                        self.df.loc[k, category.name + ":energy_consumption:hydrogen"] = 0.0

            # Mean consumption
            for k in self.df.index:
                self.df.loc[k, category.name + ":energy_consumption"] = self.df.loc[
                    k, category.name + ":energy_consumption:dropin_fuel"
                ] * (self.df.loc[k, category.name + ":share:dropin_fuel"] / 100) + self.df.loc[
                    k, category.name + ":energy_consumption:hydrogen"
                ] * (
                    self.df.loc[k, category.name + ":share:hydrogen"] / 100
                )
        # Mean energy consumption for the global fleet
        # Considering fixed category distribution in 2019
        # var_name = "global_fleet:energy_consumption"
        # self.df[var_name] = self.df["Short Range" + ":energy_consumption"] * 0.272
        # self.df[var_name] += self.df["Medium Range" + ":energy_consumption"] * 0.351
        # self.df[var_name] += self.df["Long Range" + ":energy_consumption"] * 0.377

    def _compute_mean_doc_non_energy(self):
        for category in self.fleet.categories.values():
            # Mean energy consumption per category
            # Initialization
            self.df[category.name + ":doc_non_energy:dropin_fuel"] = 0.0
            self.df[category.name + ":doc_non_energy:hydrogen"] = 0.0
            # Calculation
            for subcategory in category.subcategories.values():
                # TODO: verify aircraft order
                for k in self.df.index:
                    if self.df.loc[k, category.name + ":share:dropin_fuel"] != 0.0:
                        self.df.loc[
                            k, category.name + ":doc_non_energy:dropin_fuel"
                        ] += self.df.loc[
                            k,
                            category.name + ":" + subcategory.name + ":doc_non_energy:dropin_fuel",
                        ] / (
                            self.df.loc[k, category.name + ":share:dropin_fuel"] / 100
                        )
                    else:
                        self.df.loc[k, category.name + ":doc_non_energy:dropin_fuel"] = 0.0

                    if self.df.loc[k, category.name + ":share:hydrogen"] != 0.0:
                        self.df.loc[k, category.name + ":doc_non_energy:hydrogen"] += self.df.loc[
                            k,
                            category.name + ":" + subcategory.name + ":doc_non_energy:hydrogen",
                        ] / (self.df.loc[k, category.name + ":share:hydrogen"] / 100)
                    else:
                        self.df.loc[k, category.name + ":doc_non_energy:hydrogen"] = 0.0

            # Mean consumption
            for k in self.df.index:
                self.df.loc[k, category.name + ":doc_non_energy"] = self.df.loc[
                    k, category.name + ":doc_non_energy:dropin_fuel"
                ] * (self.df.loc[k, category.name + ":share:dropin_fuel"] / 100) + self.df.loc[
                    k, category.name + ":doc_non_energy:hydrogen"
                ] * (
                    self.df.loc[k, category.name + ":share:hydrogen"] / 100
                )

    def _compute_aircraft_share(self):
        # Aircraft share computation
        for category in self.fleet.categories.values():
            # TODO: handling of subcategory

            for key, subcategory in reversed(category.subcategories.items()):
                # TODO: verify aircraft order
                for i, aircraft in reversed(subcategory.aircraft.items()):
                    if (i == list(subcategory.aircraft.keys())[-1]) and (
                        key == list(category.subcategories.keys())[-1]
                    ):
                        aircraft_share = self.df[
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + aircraft.name
                            + ":single_aircraft_share"
                        ]

                        var_name = (
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + aircraft.name
                            + ":aircraft_share"
                        )

                        self.df[var_name] = aircraft_share

                    elif (i == list(subcategory.aircraft.keys())[-1]) and (
                        key != list(category.subcategories.keys())[-1]
                    ):
                        single_aircraft_share = self.df[
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + aircraft.name
                            + ":single_aircraft_share"
                        ]
                        single_aircraft_share_n1 = self.df[
                            category.name
                            + ":"
                            + category.subcategories[key + 1].name
                            + ":"
                            + category.subcategories[key + 1].aircraft[0].name
                            + ":single_aircraft_share"
                        ]

                        var_name = (
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + aircraft.name
                            + ":aircraft_share"
                        )
                        aircraft_share = single_aircraft_share - single_aircraft_share_n1
                        self.df[var_name] = aircraft_share
                    else:
                        single_aircraft_share = self.df[
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + aircraft.name
                            + ":single_aircraft_share"
                        ]
                        single_aircraft_share_n1 = self.df[
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + subcategory.aircraft[i + 1].name
                            + ":single_aircraft_share"
                        ]

                        var_name = (
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + aircraft.name
                            + ":aircraft_share"
                        )
                        aircraft_share = single_aircraft_share - single_aircraft_share_n1
                        self.df[var_name] = aircraft_share

            # For a category
            ref_recent_single_aircraft_share = self.df[
                category.name
                + ":"
                + category.subcategories[0].name
                + ":"
                + "recent_reference:single_aircraft_share"
            ]
            if subcategory.aircraft:
                next_aircraft_single_share = self.df[
                    category.name
                    + ":"
                    + category.subcategories[0].name
                    + ":"
                    + subcategory.aircraft[0].name
                    + ":single_aircraft_share"
                ]
            else:
                next_aircraft_single_share = 0.0
            var_name = (
                category.name
                + ":"
                + category.subcategories[0].name
                + ":"
                + "recent_reference:aircraft_share"
            )
            ref_recent_aircraft_share = (
                ref_recent_single_aircraft_share - next_aircraft_single_share
            )
            self.df[var_name] = ref_recent_aircraft_share

            # Reference old aircraft
            ref_old_aircraft_share = 100 - ref_recent_single_aircraft_share
            var_name = (
                category.name
                + ":"
                + category.subcategories[0].name
                + ":"
                + "old_reference:aircraft_share"
            )
            self.df[var_name] = ref_old_aircraft_share
        # Dedicated calculation for drop-in fuel and hydrogen
        # if aircraft.energy_type == "DROP_IN_FUEL":
        #     self.df[category.name
        #             + ":"
        #             + subcategory.name
        #             + ":share:dropin_fuel"] = aircraft_share
        #
        # if aircraft.energy_type == "HYDROGEN":
        #     self.df[category.name
        #             + ":"
        #             + subcategory.name
        #             + ":"
        #             + ":share:hydrogen"] = aircraft_share

    def _compute_single_aircraft_share(self):
        # Single aircraft share computation (for obtaining the main plot on fleet renewal)
        for category in self.fleet.categories.values():

            limit = 2
            life_base = 25
            parameter_base = np.log(100 / limit - 1) / (life_base / 2)
            parameter_renewal = np.log(100 / limit - 1) / (category.parameters.life / 2)

            if len(category.subcategories.values()) == 1:
                subcategory = list(category.subcategories.values())[0]
                # Reference recent aircraft
                year_ref_recent_begin = (
                    subcategory.recent_reference_aircraft.entry_into_service_year
                )
                year_ref_recent_base = year_ref_recent_begin + life_base / 2
                year_ref_recent = (
                    self.prospection_start_year
                    - parameter_base
                    / parameter_renewal
                    * (self.prospection_start_year - year_ref_recent_base)
                )
                ref_recent_single_aircraft_share = self._compute(
                    float(category.parameters.life),
                    float(year_ref_recent),
                    # Should always be 100% as there is only one subcategory
                    float(subcategory.parameters.share),
                    recent=True,
                )
                var_name = (
                    category.name
                    + ":"
                    + subcategory.name
                    + ":"
                    + "recent_reference:single_aircraft_share"
                )
                self.df[var_name] = ref_recent_single_aircraft_share

                # Reference old aircraft
                ref_old_single_aircraft_share = 100
                var_name = (
                    category.name
                    + ":"
                    + subcategory.name
                    + ":"
                    + "old_reference:single_aircraft_share"
                )
                self.df[var_name] = ref_old_single_aircraft_share

                # New aircraft
                for aircraft in subcategory.aircraft.values():
                    single_aircraft_share = self._compute(
                        float(category.parameters.life),
                        float(aircraft.parameters.entry_into_service_year),
                        # Should always be 100% as there is only one subcategory
                        float(subcategory.parameters.share),
                    )
                    var_name = (
                        category.name
                        + ":"
                        + subcategory.name
                        + ":"
                        + aircraft.name
                        + ":single_aircraft_share"
                    )
                    self.df[var_name] = single_aircraft_share

            elif len(category.subcategories.values()) == 2:

                # We start with last subcategory
                subcategory = list(category.subcategories.values())[-1]

                for i, aircraft in subcategory.aircraft.items():
                    single_aircraft_share = self._compute(
                        float(category.parameters.life),
                        float(aircraft.parameters.entry_into_service_year),
                        float(subcategory.parameters.share),
                    )

                    var_name = (
                        category.name
                        + ":"
                        + subcategory.name
                        + ":"
                        + aircraft.name
                        + ":single_aircraft_share"
                    )
                    self.df[var_name] = single_aircraft_share
                    if i == 0:
                        oldest_single_aircraft_share = single_aircraft_share

                # We now use the first subcategory
                subcategory = list(category.subcategories.values())[0]

                # Reference recent aircraft
                year_ref_recent_begin = (
                    subcategory.recent_reference_aircraft.entry_into_service_year
                )
                year_ref_recent_base = year_ref_recent_begin + life_base / 2
                year_ref_recent = (
                    self.prospection_start_year
                    - parameter_base
                    / parameter_renewal
                    * (self.prospection_start_year - year_ref_recent_base)
                )
                ref_recent_single_aircraft_share = oldest_single_aircraft_share + self._compute(
                    float(category.parameters.life),
                    float(year_ref_recent),
                    100 - oldest_single_aircraft_share,
                    recent=True,
                )
                var_name = (
                    category.name
                    + ":"
                    + subcategory.name
                    + ":"
                    + "recent_reference:single_aircraft_share"
                )
                self.df[var_name] = ref_recent_single_aircraft_share

                # Reference old aircraft
                ref_old_single_aircraft_share = 100
                var_name = (
                    category.name
                    + ":"
                    + subcategory.name
                    + ":"
                    + "old_reference:single_aircraft_share"
                )
                self.df[var_name] = ref_old_single_aircraft_share
                # New aircraft
                for aircraft in subcategory.aircraft.values():
                    single_aircraft_share = oldest_single_aircraft_share + self._compute(
                        float(category.parameters.life),
                        float(aircraft.parameters.entry_into_service_year),
                        100 - oldest_single_aircraft_share,
                    )
                    var_name = (
                        category.name
                        + ":"
                        + subcategory.name
                        + ":"
                        + aircraft.name
                        + ":single_aircraft_share"
                    )
                    self.df[var_name] = single_aircraft_share

            # If more than two subcategories
            else:
                for key, subcategory in reversed(category.subcategories.items()):
                    # Last subcategory
                    if key == list(category.subcategories.keys())[-1]:
                        for i, aircraft in subcategory.aircraft.items():
                            single_aircraft_share = self._compute(
                                float(category.parameters.life),
                                float(aircraft.parameters.entry_into_service_year),
                                float(subcategory.parameters.share),
                            )

                            var_name = (
                                category.name
                                + ":"
                                + subcategory.name
                                + ":"
                                + aircraft.name
                                + ":single_aircraft_share"
                            )
                            self.df[var_name] = single_aircraft_share
                            if i == 0:
                                oldest_single_aircraft_share = single_aircraft_share
                    # Initial subcategory
                    elif key == list(category.subcategories.keys())[0]:
                        # Reference recent aircraft
                        year_ref_recent_begin = (
                            subcategory.recent_reference_aircraft.entry_into_service_year
                        )
                        year_ref_recent_base = year_ref_recent_begin + life_base / 2
                        year_ref_recent = (
                            self.prospection_start_year
                            - parameter_base
                            / parameter_renewal
                            * (self.prospection_start_year - year_ref_recent_base)
                        )
                        ref_recent_single_aircraft_share = (
                            oldest_single_aircraft_share
                            + self._compute(
                                float(category.parameters.life),
                                float(year_ref_recent),
                                100 - oldest_single_aircraft_share,
                                recent=True,
                            )
                        )

                        var_name = (
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + "recent_reference:single_aircraft_share"
                        )
                        self.df[var_name] = ref_recent_single_aircraft_share

                        # Reference old aircraft
                        ref_old_single_aircraft_share = 100
                        var_name = (
                            category.name
                            + ":"
                            + subcategory.name
                            + ":"
                            + "old_reference:single_aircraft_share"
                        )
                        self.df[var_name] = ref_old_single_aircraft_share

                        # New aircraft
                        for aircraft in subcategory.aircraft.values():
                            single_aircraft_share = oldest_single_aircraft_share + self._compute(
                                float(category.parameters.life),
                                float(aircraft.parameters.entry_into_service_year),
                                100 - oldest_single_aircraft_share,
                            )
                            var_name = (
                                category.name
                                + ":"
                                + subcategory.name
                                + ":"
                                + aircraft.name
                                + ":single_aircraft_share"
                            )
                            self.df[var_name] = single_aircraft_share

                    else:
                        for i, aircraft in subcategory.aircraft.items():
                            single_aircraft_share = oldest_single_aircraft_share + self._compute(
                                float(category.parameters.life),
                                float(aircraft.parameters.entry_into_service_year),
                                float(subcategory.parameters.share),
                            )
                            var_name = (
                                category.name
                                + ":"
                                + subcategory.name
                                + ":"
                                + aircraft.name
                                + ":single_aircraft_share"
                            )
                            self.df[var_name] = single_aircraft_share
                            if i == 0:
                                new_oldest_single_aircraft_share = single_aircraft_share
                        oldest_single_aircraft_share = new_oldest_single_aircraft_share
        return subcategory

    def plot(self):
        x = np.linspace(
            self.prospection_start_year,
            self.end_year,
            self.end_year - self.prospection_start_year + 1,
        )

        categories = list(self.fleet.categories.values())

        f, axs = plt.subplots(2, len(categories), figsize=(20, 10))

        for i, category in enumerate(categories):
            # Top plot
            ax = axs[0, i]

            # Initial subcategory
            subcategory = category.subcategories[0]
            # Old reference aircraft
            var_name = (
                category.name + ":" + subcategory.name + ":" + "old_reference:single_aircraft_share"
            )
            ax.fill_between(
                x,
                self.df.loc[self.prospection_start_year : self.end_year, var_name],
                label=subcategory.name + " - Old reference aircraft",
            )

            # Recent reference aircraft
            var_name = (
                category.name
                + ":"
                + subcategory.name
                + ":"
                + "recent_reference:single_aircraft_share"
            )
            ax.fill_between(
                x,
                self.df.loc[self.prospection_start_year : self.end_year, var_name],
                label=subcategory.name + " - Recent reference aircraft",
            )

            for j, subcategory in category.subcategories.items():

                # New aircraft
                for aircraft in subcategory.aircraft.values():
                    var_name = (
                        category.name
                        + ":"
                        + subcategory.name
                        + ":"
                        + aircraft.name
                        + ":single_aircraft_share"
                    )
                    ax.fill_between(
                        x,
                        self.df.loc[self.prospection_start_year : self.end_year, var_name],
                        label=subcategory.name + " - " + aircraft.name,
                    )

            ax.set_xlim(self.prospection_start_year, self.end_year)
            ax.set_ylim(0, 100)
            ax.legend(loc="upper left", prop={"size": 8})
            ax.set_xlabel("Year")
            ax.set_ylabel("Share in fleet [%]")
            ax.set_title(categories[i].name)

            # Bottom plot
            ax = axs[1, i]
            ax.plot(
                x,
                self.df.loc[
                    self.prospection_start_year : self.end_year,
                    category.name + ":energy_consumption",
                ],
            )

            ax.set_xlim(self.prospection_start_year, self.end_year)
            ax.set_xlabel("Year")
            ax.set_ylabel("Fleet mean energy consumption [MJ/ASK]")
            # ax.set_title(categories[i].name)

        plt.plot()
        # plt.savefig("fleet_renewal.pdf")

    def _compute(self, life, entry_into_service_year, share, recent=False):
        x = np.linspace(
            self.historic_start_year, self.end_year, self.end_year - self.historic_start_year + 1
        )

        # Intermediate variable for S-shaped function
        limit = 2
        growth_rate = np.log(100 / limit - 1) / (life / 2)

        if not recent:
            midpoint_year = entry_into_service_year + life / 2
        else:
            midpoint_year = entry_into_service_year

        y_share = share / (1 + np.exp(-growth_rate * (x - midpoint_year)))
        y_share_max = 100 / (1 + np.exp(-growth_rate * (x - midpoint_year)))

        y = np.where(y_share_max < limit, 0.0, y_share)
        return y


class Aircraft(object):
    def __init__(
        self,
        name: str = None,
        parameters: AircraftParameters = None,
        energy_type="DROP_IN_FUEL",
    ):
        self.name = name
        if parameters is None:
            parameters = AircraftParameters()
        self.parameters = parameters
        self.energy_type = energy_type

    def from_dataframe_row(self, row):
        self.name = row[AIRCRAFT_COLUMNS[0]]
        self.parameters.entry_into_service_year = row[AIRCRAFT_COLUMNS[1]]
        self.parameters.consumption_evolution = row[AIRCRAFT_COLUMNS[2]]
        self.parameters.nox_evolution = row[AIRCRAFT_COLUMNS[3]]
        self.parameters.soot_evolution = row[AIRCRAFT_COLUMNS[4]]
        self.parameters.doc_non_energy_evolution = row[AIRCRAFT_COLUMNS[5]]
        self.parameters.cruise_altitude = row[AIRCRAFT_COLUMNS[6]]
        self.energy_type = row[AIRCRAFT_COLUMNS[7]]

        return self


class SubCategory(object):
    def __init__(self, name: str = None, parameters: SubcategoryParameters = None):
        self.name = name
        if parameters is None:
            parameters = SubcategoryParameters()
        self.parameters = parameters
        self.aircraft = {}
        self.old_reference_aircraft = ReferenceAircraftParameters()
        self.recent_reference_aircraft = ReferenceAircraftParameters()

        self._setup_datagrid()

        self._setup_ui()

    def add_aircraft(self, change=None, aircraft: Aircraft = None):

        if aircraft is None:
            aircraft_data = np.array(
                [
                    "New aircraft",
                    2025,
                    5.0,
                    10.0,
                    10.0,
                    12000.0,
                    "DROP_IN_FUEL",
                ]
            ).reshape((1, len(AIRCRAFT_COLUMNS)))
        else:
            aircraft_data = np.array(
                [
                    aircraft.name,
                    aircraft.parameters.entry_into_service_year,
                    aircraft.parameters.consumption_evolution,
                    aircraft.parameters.nox_evolution,
                    aircraft.parameters.soot_evolution,
                    aircraft.parameters.doc_non_energy_evolution,
                    aircraft.parameters.cruise_altitude,
                    aircraft.energy_type,
                ]
            ).reshape((1, len(AIRCRAFT_COLUMNS)))

        # Add aircraft to grid
        current_grid_df = self.datagrid.data
        if len(current_grid_df) == 0:
            new_index = 0
        else:
            new_index = current_grid_df.index[-1] + 1

        current_grid_df = self.datagrid.data
        additional_row = pd.DataFrame(
            columns=current_grid_df.columns,
            index=[new_index],
            data=aircraft_data,
        )

        self.datagrid.data = pd.concat([self.datagrid.data, additional_row], ignore_index=True)

        # TODO: see if we avoid this when aircraft is directly provided
        self._update_parameters_from_grid()

    def remove_aircraft(self, change=None, aircraft_name: str = None):

        current_grid_df = self.datagrid.data

        if aircraft_name is not None:
            index_to_remove = current_grid_df.index[
                current_grid_df[AIRCRAFT_COLUMNS[0]] == aircraft_name
            ].tolist()
        else:
            selected_cells = self.datagrid.selected_cells

            index_to_remove = []
            for cell in selected_cells:
                row_index = cell["r"]
                if row_index not in index_to_remove:
                    index_to_remove.append(row_index)

        self.datagrid.data = current_grid_df.drop(index=index_to_remove)

        self._update_parameters_from_grid()

    def compute(self):
        for aircraft in self.aircraft:
            aircraft.compute()

    def from_dataframe_row(self, row):
        self.name = row[SUBCATEGORY_COLUMNS[0]]
        self.parameters.share = row[SUBCATEGORY_COLUMNS[1]]

        return self

    def _setup_datagrid(self):

        df = pd.DataFrame(columns=AIRCRAFT_COLUMNS)

        self.datagrid = dg.DataGrid(df, selection_mode="cell", editable=True)
        self.datagrid.auto_fit_columns = True

    def _setup_ui(self):
        button_add_row = ipw.Button(
            description="Add Aircraft", style=ipw.ButtonStyle(button_color="darkgreen")
        )
        button_remove_row = ipw.Button(
            description="Remove Aircraft", style=ipw.ButtonStyle(button_color="darkred")
        )

        button_add_row.on_click(self.add_aircraft)
        button_remove_row.on_click(self.remove_aircraft)

        self.datagrid.on_cell_change(self._update_parameters_from_grid)

        self.ui = ipw.VBox([ipw.HBox([button_add_row, button_remove_row]), self.datagrid])
        self.ui.layout.width = "1200px"
        self.datagrid.auto_fit_columns = True

    def _update_parameters_from_grid(self, change=None):
        self.aircraft = {}
        current_grid_df = self.datagrid.data

        for index, row in current_grid_df.iterrows():
            aircraft = Aircraft()
            # self.aircraft[row["Name"]] = aircraft.from_dataframe_row(row)
            self.aircraft[int(index)] = aircraft.from_dataframe_row(row)


class Category(object):
    def __init__(self, name, parameters: CategoryParameters):
        self.name = name
        self.parameters = parameters
        self.subcategories = {}
        self.total_shares = 100

        self._setup_datagrid()

        self._setup_ui()

    def _compute(self):
        self._check_shares()
        if self.subcategories:
            for subcategory in self.subcategories.values():
                subcategory.compute()

    def add_subcategory(self, change=None, subcategory: SubCategory = None):

        if subcategory is None:
            subcategory_data = np.array(
                [
                    "New subcategory",
                    0.0,
                ]
            ).reshape((1, len(SUBCATEGORY_COLUMNS)))
        else:
            subcategory_data = np.array(
                [
                    subcategory.name,
                    subcategory.parameters.share,
                ]
            ).reshape((1, len(SUBCATEGORY_COLUMNS)))

        # Add subcategory to grid
        current_grid_df = self.datagrid.data
        if len(current_grid_df) == 0:
            new_index = 0
        else:
            new_index = current_grid_df.index[-1] + 1

        additional_row = pd.DataFrame(
            columns=current_grid_df.columns,
            index=[new_index],
            data=subcategory_data,
        )

        if subcategory is not None:
            self.subcategories[new_index] = subcategory
        else:
            # TODO: can we really have several rows?
            for index, row in additional_row.iterrows():
                subcategory = SubCategory()
                self.subcategories[new_index] = subcategory.from_dataframe_row(row)

        self.datagrid.data = pd.concat([self.datagrid.data, additional_row], ignore_index=True)
        self.datagrid.auto_fit_columns = True

    def remove_subcategory(self, change=None, subcategory_name: str = None):

        current_grid_df = self.datagrid.get_visible_data()

        if subcategory_name is not None:
            index_to_remove = current_grid_df.index[
                current_grid_df[SUBCATEGORY_COLUMNS[0]] == subcategory_name
            ].tolist()
        else:
            selected_cells = self.datagrid.selected_cells

            index_to_remove = []
            for cell in selected_cells:
                row_index = cell["r"]
                if row_index not in index_to_remove:
                    index_to_remove.append(row_index)

        # Remove and reset index
        for index in index_to_remove:
            self.subcategories.pop(index)
        self.subcategories = {
            i: self.subcategories[k] for i, k in enumerate(sorted(self.subcategories.keys()))
        }
        self.datagrid.data = current_grid_df.drop(index=index_to_remove)
        self.datagrid.auto_fit_columns = True

    def _setup_datagrid(self):

        df = pd.DataFrame(columns=SUBCATEGORY_COLUMNS)

        self.datagrid = dg.DataGrid(df, selection_mode="cell", editable=True)
        self.datagrid.auto_fit_columns = True
        self.datagrid.on_cell_change(self._update)

    def _setup_ui(self):
        self.button_add_row = ipw.Button(
            description="Add Subcategory", style=ipw.ButtonStyle(button_color="darkgreen")
        )
        self.button_remove_row = ipw.Button(
            description="Remove Subcategory", style=ipw.ButtonStyle(button_color="darkred")
        )

        self.button_add_row.on_click(self.add_subcategory)
        self.button_remove_row.on_click(self.remove_subcategory)
        self.datagrid.on_cell_change(self._update)

        self.ui = ipw.VBox([ipw.HBox([self.button_add_row, self.button_remove_row]), self.datagrid])
        self.ui.layout.width = "600px"

    def _update(self, change=None):
        self._update_parameters_from_grid()
        # self.total_shares = np.sum(
        #     [share for share in self.subcategories.values().parameters.share]
        # )
        for subcategory in self.subcategories.values():
            subcategory._update_parameters_from_grid()

    def _check_shares(self):
        self._update()
        if not np.isclose(self.total_shares, 100.0, atol=0.1):
            raise UserWarning(
                "Total shares for category %s are %f instead of 100%",
                (self.name, self.total_shares),
            )

    def _update_parameters_from_grid(self):
        current_grid_df = self.datagrid.data

        for index, row in current_grid_df.iterrows():
            self.subcategories[index].from_dataframe_row(row)


class Fleet(object):
    def __init__(
        self,
        add_examples_aircraft_and_subcategory=True,
    ):
        self.categories = {}

        # Build default fleet
        self._build_default_fleet(
            add_examples_aircraft_and_subcategory=add_examples_aircraft_and_subcategory
        )

        # Initialize
        self.ui = None

        # Setup tree
        self.tree = Tree(stripes=True)
        self.tree.observe(self._update)
        self._setup_tree()

        # Take first category as default
        self.selected_item = list(self.categories.values())[0]

        # Setup user interface
        self._setup_ui()

    def compute(self):
        for cat in self.categories.values():
            cat._compute()

    def _setup_tree(self):
        for name, category in self.categories.items():
            category.button_add_row.observe(self._update)
            category.button_remove_row.observe(self._update)
            node = Node(name)
            subcategory_names = []
            for name, subcategory in category.subcategories.items():
                subcategory_names.append(subcategory.name)
            for name in subcategory_names:
                subnode = Node(name)
                node.add_node(subnode)
            self.tree.add_node(node)

    def _update_tree(self):
        for cat_node in list(self.tree.nodes):
            subcat_nodes = list(cat_node.nodes)
            subcategory_names = [
                subcategory.name
                for index, subcategory in self.categories[cat_node.name].subcategories.items()
            ]
            # Remove old nodes
            for node in subcat_nodes:
                if node.name not in subcategory_names:
                    cat_node.remove_node(node)

            subcategory_node_names = [node.name for node in list(cat_node.nodes)]

            # Add new nodes
            for name in subcategory_names:
                if name not in subcategory_node_names:
                    subnode = Node(name)
                    cat_node.add_node(subnode)

    def _update(self, change=None):
        if self.tree.selected_nodes:
            selected_node_name = list(self.tree.selected_nodes)[0].name
        else:
            selected_node_name = self.tree.nodes[0].name

        self.selected_item = self._find_category_or_subcategory(selected_node_name)
        self._update_tree()
        self._update_ui()

    def _find_category_or_subcategory(self, selected_node_name):
        for name, category in self.categories.items():
            if category.name == selected_node_name:
                return category
            else:
                for index, subcategory in category.subcategories.items():
                    if subcategory.name == selected_node_name:
                        return subcategory

    def _setup_ui(self):
        self.ui = ipw.HBox([self.tree, self.selected_item.ui])
        self.ui.layout

    def _update_ui(self):
        if self.ui is not None:
            self.ui.children = (self.tree, self.selected_item.ui)
            if isinstance(self.selected_item, Category):
                self.selected_item.button_add_row.on_click(self._update)
                self.selected_item.button_remove_row.on_click(self._update)
                self.selected_item.datagrid.on_cell_change(self._update)

    def _build_default_fleet(self, add_examples_aircraft_and_subcategory=True):
        # Short range narrow-body
        aircraft_params = AircraftParameters(
            entry_into_service_year=2035,
            consumption_evolution=-15.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        sr_nb_aircraft_1 = Aircraft(
            "New Short-range Narrow-body 1", parameters=aircraft_params, energy_type="DROP_IN_FUEL"
        )

        aircraft_params = AircraftParameters(
            entry_into_service_year=2045,
            consumption_evolution=-30.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        sr_nb_aircraft_2 = Aircraft(
            "New Short-range Narrow-body 2", parameters=aircraft_params, energy_type="DROP_IN_FUEL"
        )

        # Short range regional turboprop
        aircraft_params = AircraftParameters(
            entry_into_service_year=2030,
            consumption_evolution=-20.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=6000.0,
        )

        sr_tp_aircraft_1 = Aircraft(
            "New Regional turboprop 1",
            parameters=aircraft_params,
            energy_type="DROP_IN_FUEL",
        )

        aircraft_params = AircraftParameters(
            entry_into_service_year=2045,
            consumption_evolution=-35.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=6000.0,
        )

        sr_tp_aircraft_2 = Aircraft(
            "New Regional turboprop 2",
            parameters=aircraft_params,
            energy_type="DROP_IN_FUEL",
        )

        # Short range regional turbofan
        aircraft_params = AircraftParameters(
            entry_into_service_year=2035,
            consumption_evolution=-15.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        sr_tf_aircraft_1 = Aircraft(
            "New Regional turbofan 1",
            parameters=aircraft_params,
            energy_type="DROP_IN_FUEL",
        )

        aircraft_params = AircraftParameters(
            entry_into_service_year=2045,
            consumption_evolution=-30.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        sr_tf_aircraft_2 = Aircraft(
            "New Regional turbofan 2",
            parameters=aircraft_params,
            energy_type="DROP_IN_FUEL",
        )

        # Short range hydrogen aircraft

        aircraft_params = AircraftParameters(
            entry_into_service_year=2035,
            consumption_evolution=10.0,
            nox_evolution=0.0,
            soot_evolution=-100.0,
            doc_non_energy_evolution=10.0,
            cruise_altitude=12000.0,
        )

        sr_aircraft_hydrogen = Aircraft(
            "New Short-range hydrogen", parameters=aircraft_params, energy_type="HYDROGEN"
        )

        # Medium range
        aircraft_params = AircraftParameters(
            entry_into_service_year=2035,
            consumption_evolution=-15.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        mr_aircraft_1 = Aircraft(
            "New Medium-range narrow-body 1", parameters=aircraft_params, energy_type="DROP_IN_FUEL"
        )

        aircraft_params = AircraftParameters(
            entry_into_service_year=2045,
            consumption_evolution=-30.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        mr_aircraft_2 = Aircraft(
            "New Medium-range narrow-body 2", parameters=aircraft_params, energy_type="DROP_IN_FUEL"
        )

        # Long range
        aircraft_params = AircraftParameters(
            entry_into_service_year=2035,
            consumption_evolution=-15.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        lr_aircraft_1 = Aircraft(
            "New Long-range wide-body 1", parameters=aircraft_params, energy_type="DROP_IN_FUEL"
        )

        aircraft_params = AircraftParameters(
            entry_into_service_year=2045,
            consumption_evolution=-30.0,
            nox_evolution=0.0,
            soot_evolution=0.0,
            doc_non_energy_evolution=0.0,
            cruise_altitude=12000.0,
        )

        lr_aircraft_2 = Aircraft(
            "New Long-range wide-body 2", parameters=aircraft_params, energy_type="DROP_IN_FUEL"
        )

        # Short range narrow-body
        if add_examples_aircraft_and_subcategory:
            subcat_params = SubcategoryParameters(share=20.0)
        else:
            subcat_params = SubcategoryParameters(share=100.0)
        sr_nb_cat = SubCategory("SR conventional narrow-body", parameters=subcat_params)
        # Reference aircraft
        # Old
        sr_nb_cat.old_reference_aircraft.entry_into_service_year = 1970
        sr_nb_cat.old_reference_aircraft.energy_per_ask = 110.8 / 73.2 * 0.824  # [MJ/ASK]
        sr_nb_cat.old_reference_aircraft.emission_index_nox = 0.01514
        sr_nb_cat.old_reference_aircraft.emission_index_soot = 3e-5
        sr_nb_cat.old_reference_aircraft.doc_non_energy_base = 0.045
        sr_nb_cat.old_reference_aircraft.cruise_altitude = 12000.0

        # Recent
        sr_nb_cat.recent_reference_aircraft.entry_into_service_year = 2007.13
        sr_nb_cat.recent_reference_aircraft.energy_per_ask = 84.2 / 73.2 * 0.824  # [MJ/ASK]
        sr_nb_cat.recent_reference_aircraft.emission_index_nox = 0.01514
        sr_nb_cat.recent_reference_aircraft.emission_index_soot = 3e-5
        sr_nb_cat.recent_reference_aircraft.doc_non_energy_base = 0.045
        sr_nb_cat.recent_reference_aircraft.cruise_altitude = 12000.0

        if add_examples_aircraft_and_subcategory:
            sr_nb_cat.add_aircraft(aircraft=sr_nb_aircraft_1)
            sr_nb_cat.add_aircraft(aircraft=sr_nb_aircraft_2)

        # Short range hydrogen aircraft
        if add_examples_aircraft_and_subcategory:
            subcat_params = SubcategoryParameters(share=50.0)
            sr_subcat_hydrogen = SubCategory(
                "SR hydrogen conventional narrow-body", parameters=subcat_params
            )

        if add_examples_aircraft_and_subcategory:
            sr_subcat_hydrogen.add_aircraft(aircraft=sr_aircraft_hydrogen)

        # Short range regional turboprop
        if add_examples_aircraft_and_subcategory:
            subcat_params = SubcategoryParameters(share=30.0)
            sr_rp_cat = SubCategory("SR regional turboprop", parameters=subcat_params)
        # Reference aircraft
        # Old
        # sr_rp_cat.old_reference_aircraft.entry_into_service_year = 1970
        # sr_rp_cat.old_reference_aircraft.energy_per_ask = 101.2 / 73.2 * 0.824  # [MJ/ASK]
        # sr_rp_cat.old_reference_aircraft.emission_index_nox = 0.01514
        # sr_rp_cat.old_reference_aircraft.emission_index_soot = 3e-5
        # sr_rp_cat.old_reference_aircraft.cruise_altitude = 6000.0

        # Recent
        # sr_rp_cat.recent_reference_aircraft.entry_into_service_year = 2005
        # sr_rp_cat.recent_reference_aircraft.energy_per_ask = 101.2 / 73.2 * 0.824  # [MJ/ASK]
        # sr_rp_cat.recent_reference_aircraft.emission_index_nox = 0.01514
        # sr_rp_cat.recent_reference_aircraft.emission_index_soot = 3e-5
        # sr_rp_cat.recent_reference_aircraft.cruise_altitude = 6000.0

        if add_examples_aircraft_and_subcategory:
            sr_rp_cat.add_aircraft(aircraft=sr_tp_aircraft_1)
            sr_rp_cat.add_aircraft(aircraft=sr_tp_aircraft_2)

        # Short range regional turbofan
        subcat_params = SubcategoryParameters(share=0.0)
        sr_tf_cat = SubCategory("SR regional turbofan", parameters=subcat_params)
        # Reference aircraft
        # Old
        # sr_tf_cat.old_reference_aircraft.entry_into_service_year = 1970
        # sr_tf_cat.old_reference_aircraft.energy_per_ask = 192.9 / 73.2 * 0.824  # [MJ/ASK]
        # sr_tf_cat.old_reference_aircraft.emission_index_nox = 0.01514
        # sr_tf_cat.old_reference_aircraft.emission_index_soot = 3e-5
        # sr_tf_cat.old_reference_aircraft.cruise_altitude = 12000.0

        # Recent
        # sr_tf_cat.recent_reference_aircraft.entry_into_service_year = 2000
        # sr_tf_cat.recent_reference_aircraft.energy_per_ask = 192.9 / 73.2 * 0.824  # [MJ/ASK]
        # sr_tf_cat.recent_reference_aircraft.emission_index_nox = 0.01514
        # sr_tf_cat.recent_reference_aircraft.emission_index_soot = 3e-5
        # sr_tf_cat.recent_reference_aircraft.cruise_altitude = 12000.0

        # sr_tf_cat.add_aircraft(aircraft=sr_tf_aircraft_1)
        # sr_tf_cat.add_aircraft(aircraft=sr_tf_aircraft_2)

        # Short range
        cat_params = CategoryParameters(life=25)
        sr_cat = Category("Short Range", parameters=cat_params)
        sr_cat.add_subcategory(subcategory=sr_nb_cat)
        if add_examples_aircraft_and_subcategory:
            sr_cat.add_subcategory(subcategory=sr_rp_cat)
            sr_cat.add_subcategory(subcategory=sr_subcat_hydrogen)
        # sr_cat.add_subcategory(subcategory=sr_tf_cat)

        # Medium range
        subcat_params = SubcategoryParameters(share=100.0)
        mr_subcat = SubCategory("MR conventional narrow-body", parameters=subcat_params)
        # Reference aircraft
        # Old
        mr_subcat.old_reference_aircraft.entry_into_service_year = 1970
        mr_subcat.old_reference_aircraft.energy_per_ask = 81.4 / 73.2 * 0.824  # [MJ/ASK]
        mr_subcat.old_reference_aircraft.emission_index_nox = 0.01514
        mr_subcat.old_reference_aircraft.emission_index_soot = 3e-5
        mr_subcat.old_reference_aircraft.doc_non_energy_base = 0.028
        mr_subcat.old_reference_aircraft.cruise_altitude = 12000.0

        # Recent
        mr_subcat.recent_reference_aircraft.entry_into_service_year = 2010.35
        mr_subcat.recent_reference_aircraft.energy_per_ask = 62.0 / 73.2 * 0.824  # [MJ/ASK]
        mr_subcat.recent_reference_aircraft.emission_index_nox = 0.01514
        mr_subcat.recent_reference_aircraft.emission_index_soot = 3e-5
        mr_subcat.recent_reference_aircraft.doc_non_energy_base = 0.028
        mr_subcat.recent_reference_aircraft.cruise_altitude = 12000.0

        if add_examples_aircraft_and_subcategory:
            mr_subcat.add_aircraft(aircraft=mr_aircraft_1)
            mr_subcat.add_aircraft(aircraft=mr_aircraft_2)

        cat_params = CategoryParameters(life=25)
        mr_cat = Category(name="Medium Range", parameters=cat_params)
        mr_cat.add_subcategory(subcategory=mr_subcat)

        # Long range
        subcat_params = SubcategoryParameters(share=100.0)
        lr_subcat = SubCategory("LR conventional wide-body", parameters=subcat_params)
        # Reference aircraft
        # Old
        lr_subcat.old_reference_aircraft.entry_into_service_year = 1970
        lr_subcat.old_reference_aircraft.energy_per_ask = 96.65 / 73.2 * 0.824  # [MJ/ASK]
        lr_subcat.old_reference_aircraft.emission_index_nox = 0.01514
        lr_subcat.old_reference_aircraft.emission_index_soot = 3e-5
        lr_subcat.old_reference_aircraft.doc_non_energy_base = 0.023
        lr_subcat.old_reference_aircraft.cruise_altitude = 12000.0

        # Recent
        lr_subcat.recent_reference_aircraft.entry_into_service_year = 2009.36
        lr_subcat.recent_reference_aircraft.energy_per_ask = 73.45 / 73.2 * 0.824  # [MJ/ASK]
        lr_subcat.recent_reference_aircraft.emission_index_nox = 0.01514
        lr_subcat.recent_reference_aircraft.emission_index_soot = 3e-5
        lr_subcat.recent_reference_aircraft.doc_non_energy_base = 0.023
        lr_subcat.recent_reference_aircraft.cruise_altitude = 12000.0

        if add_examples_aircraft_and_subcategory:
            lr_subcat.add_aircraft(aircraft=lr_aircraft_1)
            lr_subcat.add_aircraft(aircraft=lr_aircraft_2)

        cat_params = CategoryParameters(life=25)
        lr_cat = Category("Long Range", parameters=cat_params)
        lr_cat.add_subcategory(subcategory=lr_subcat)

        self.categories[sr_cat.name] = sr_cat
        self.categories[mr_cat.name] = mr_cat
        self.categories[lr_cat.name] = lr_cat



from typing import Tuple
from scipy.optimize import fsolve


from aeromaps.models.base import AeromapsModel


class EquivalentGrossCarbonBudget(AeromapsModel):
    def __init__(self, name="equivalent_gross_carbon_budget", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        gross_carbon_budget: float = 0.0,
        T_nonCO2: float = 0.0,
        TCRE: float = 0.0,
        world_ghg_emissions_2019: float = 0.0,
        aviation_equivalentcarbonbudget_allocated_share: float = 0.0,
    ) -> Tuple[float, float, float]:
        """Gross equivalent carbon budget."""

        equivalent_gross_carbon_budget = gross_carbon_budget + T_nonCO2 / TCRE

        data = [equivalent_gross_carbon_budget, world_ghg_emissions_2019]

        average_ghg_emissions_decline_rate = float(
            fsolve(self._compute_average_ghg_emissions_decline_rate, -0.02, args=data)
        )
        equivalent_gross_carbon_budget_2050 = (
            world_ghg_emissions_2019
            * (
                (1 - average_ghg_emissions_decline_rate)
                - (1 - average_ghg_emissions_decline_rate) ** 32
            )
            / average_ghg_emissions_decline_rate
        )

        aviation_equivalent_carbon_budget = (
            aviation_equivalentcarbonbudget_allocated_share
            / 100
            * equivalent_gross_carbon_budget_2050
        )

        self.float_outputs["equivalent_gross_carbon_budget"] = equivalent_gross_carbon_budget
        self.float_outputs[
            "equivalent_gross_carbon_budget_2050"
        ] = equivalent_gross_carbon_budget_2050
        self.float_outputs["aviation_equivalent_carbon_budget"] = aviation_equivalent_carbon_budget

        return (
            equivalent_gross_carbon_budget,
            equivalent_gross_carbon_budget_2050,
            aviation_equivalent_carbon_budget,
        )

    @staticmethod
    def _compute_average_ghg_emissions_decline_rate(x, data):
        equivalent_gross_carbon_budget, world_ghg_emissions_2019 = data
        return (
            (1 - x) - (1 - x) ** 82
        ) / x - equivalent_gross_carbon_budget / world_ghg_emissions_2019



from dataclasses import dataclass


@dataclass
class SustainabilityAssessmentParameters(object):

    # Carbon budgets
    net_carbon_budget: float = 900
    T_nonCO2: float = 0.2
    carbon_dioxyde_removal_2100: float = 100
    world_co2_emissions_2019: float = 43.05
    world_ghg_emissions_2019: float = 59
    aviation_carbon_budget_allocated_share: float = 2.6
    aviation_equivalentcarbonbudget_allocated_share: float = 5.1

    # Energy availability - Biomass
    waste_biomass: float = 12
    crops_biomass: float = 63
    forest_residues_biomass: float = 17
    agricultural_residues_biomass: float = 57
    algae_biomass: float = 15
    fog_waste_biomass: float = 1
    oil_crops_biomass_share: float = 9
    sugarystarchy_crops_biomass_share: float = 28
    lignocellulosic_crops_biomass_share: float = 63  # Model here ?
    aviation_biomass_allocated_share: float = 10

    # Energy availability - Electricity
    available_electricity: float = 300
    aviation_electricity_allocated_share: float = 3



from typing import Tuple
from aeromaps.models.base import AeromapsModel


class BiomassAvailability(AeromapsModel):
    def __init__(self, name="biomass_availability", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        waste_biomass: float = 0.0,
        crops_biomass: float = 0.0,
        forest_residues_biomass: float = 0.0,
        agricultural_residues_biomass: float = 0.0,
        algae_biomass: float = 0.0,
        fog_waste_biomass: float = 0.0,
        oil_crops_biomass_share: float = 0.0,
        sugarystarchy_crops_biomass_share: float = 0.0,
        lignocellulosic_crops_biomass_share: float = 0.0,
        aviation_biomass_allocated_share: float = 0.0,
    ) -> Tuple[float, float, float, float, float, float, float, float, float, float, float]:
        """Biomass distribution for waste and crops, Biomass availability and Biomass for aviation."""

        solid_waste_biomass = waste_biomass - fog_waste_biomass
        oil_crops_biomass = oil_crops_biomass_share / 100 * crops_biomass
        sugarystarchy_crops_biomass = sugarystarchy_crops_biomass_share / 100 * crops_biomass
        lignocellulosic_crops_biomass = lignocellulosic_crops_biomass_share / 100 * crops_biomass

        available_biomass_hefa_fog = fog_waste_biomass
        available_biomass_hefa_others = algae_biomass + oil_crops_biomass
        available_biomass_ft_others = (
            lignocellulosic_crops_biomass + forest_residues_biomass + agricultural_residues_biomass
        )
        available_biomass_ft_msw = solid_waste_biomass
        available_biomass_atj = sugarystarchy_crops_biomass

        available_biomass_total = (
            waste_biomass
            + crops_biomass
            + forest_residues_biomass
            + agricultural_residues_biomass
            + algae_biomass
        )

        aviation_available_biomass = (
            aviation_biomass_allocated_share / 100 * available_biomass_total
        )

        self.float_outputs["solid_waste_biomass"] = solid_waste_biomass
        self.float_outputs["oil_crops_biomass"] = oil_crops_biomass
        self.float_outputs["sugarystarchy_crops_biomass"] = sugarystarchy_crops_biomass
        self.float_outputs["lignocellulosic_crops_biomass"] = lignocellulosic_crops_biomass
        self.float_outputs["available_biomass_hefa_fog"] = available_biomass_hefa_fog
        self.float_outputs["available_biomass_hefa_others"] = available_biomass_hefa_others
        self.float_outputs["available_biomass_ft_others"] = available_biomass_ft_others
        self.float_outputs["available_biomass_ft_msw"] = available_biomass_ft_msw
        self.float_outputs["available_biomass_atj"] = available_biomass_atj
        self.float_outputs["available_biomass_total"] = available_biomass_total
        self.float_outputs["aviation_available_biomass"] = aviation_available_biomass

        return (
            solid_waste_biomass,
            oil_crops_biomass,
            sugarystarchy_crops_biomass,
            lignocellulosic_crops_biomass,
            available_biomass_hefa_fog,
            available_biomass_hefa_others,
            available_biomass_ft_others,
            available_biomass_ft_msw,
            available_biomass_atj,
            available_biomass_total,
            aviation_available_biomass,
        )


class ElectricityAvailability(AeromapsModel):
    def __init__(self, name="electricity_availability", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        available_electricity: float = 0.0,
        aviation_electricity_allocated_share: float = 0.0,
    ) -> float:
        """Aviation electricity availability."""
        aviation_available_electricity = (
            aviation_electricity_allocated_share / 100 * available_electricity
        )
        self.float_outputs["aviation_available_electricity"] = aviation_available_electricity

        return aviation_available_electricity
